from typing import Any, Dict, Optional, Type

from .action import ABCAction, Callback, Location, OpenLink, Text, VKApps, VKPay
from .color import KeyboardButtonColor

KEYBOARD_ACTIONS = {
    "text": Text,
    "open_link": OpenLink,
    "callback": Callback,
    "location": Location,
    "vkpay": VKPay,
    "open_app": VKApps,
}


class KeyboardButton:
    def __init__(self, action: "ABCAction", color: Optional["KeyboardButtonColor"] = None):
        if not isinstance(action, ABCAction):
            msg = "action must be instance of ABCAction"
            raise TypeError(msg)
        if color and not isinstance(color, KeyboardButtonColor):
            msg = "color must be instance of KeyboardButtonColor"
            raise TypeError(msg)
        self.action = action
        self.color = color

    @classmethod
    def from_typed(
        cls: Type["KeyboardButton"],
        action: "ABCAction",
        color: Optional["KeyboardButtonColor"] = None,
    ) -> "KeyboardButton":
        return cls(action, color)

    @classmethod
    def from_dict(cls: Type["KeyboardButton"], data: dict) -> "KeyboardButton":
        action_type = KEYBOARD_ACTIONS.get(data.pop("type", None))
        color = data.pop("color", None)
        if color:
            color = KeyboardButtonColor(color)
        if action_type is None:
            msg = "KeyboardButton action type is not defined"
            raise ValueError(msg)

        return cls(action_type(**data), color)

    def get_data(self) -> dict:
        data: Dict[str, Any] = {"action": self.action.get_data()}
        if self.action.type in ("text", "callback") and self.color:
            data["color"] = self.color.value
        return data



from typing import Any, List, Optional

from vkbottle.modules import logger

from .code_exception import CodeException
from .reducible_kwargs_exception import ReducibleKwargsException


class VKAPIError(CodeException, ReducibleKwargsException):
    def __init__(
        self, *, error_msg: str, request_params: Optional[List[dict]] = None, **kwargs: Any
    ):
        self.error_msg = error_msg
        # server maybe return error code without request_params key
        if request_params:
            self.request_params = {item["key"]: item["value"] for item in request_params}
        else:
            self.request_params = {}
        if kwargs:
            logger.warning("VK API Error {} has extra kwargs: {}", self.code, kwargs)
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.error_msg


class CaptchaError(VKAPIError, code=14):
    def __init__(
        self,
        *,
        captcha_sid: str,
        captcha_img: str,
        captcha_ts: Optional[str] = None,
        captcha_attempt: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.captcha_sid = int(captcha_sid)
        self.captcha_img = captcha_img
        self.captcha_ts = captcha_ts
        self.captcha_attempt = captcha_attempt


class APIAuthError(VKAPIError, code=5):
    def __init__(self, *, ban_info: Optional[dict] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.ban_info = ban_info



import asyncio
import inspect
import sys
import warnings
from typing import Protocol

from choicelib import choice_in_order

try:
    import pydantic.v1 as pydantic  # type: ignore
except ImportError:
    import pydantic  # type: ignore  # noqa: F401


class JSONModule(Protocol):
    def loads(self, s: str) -> dict: ...

    def dumps(self, o: dict) -> str: ...

    def load(self, f: str) -> dict: ...

    def dump(self, o: dict, f: str) -> None: ...


json: JSONModule = choice_in_order(
    ["orjson", "ujson", "hyperjson"],
    do_import=True,
    default="json",
)

warnings.simplefilter("always", DeprecationWarning)
showwarning_ = warnings.showwarning


def showwarning(message, category, filename, lineno, file=None, line=None):  # noqa: ARG001
    new_message = f"{category.__name__}: {message}"
    if logging_module == "loguru":
        logger.opt(depth=2).log("WARNING", new_message)  # type: ignore
        return
    logger.log(
        logging.WARNING,
        new_message,
        stacklevel=4,
    )


logging_module = choice_in_order(["loguru"], default="logging")
if logging_module == "loguru":
    import logging
    import os

    if not os.environ.get("LOGURU_AUTOINIT"):
        os.environ["LOGURU_AUTOINIT"] = "0"
        os.environ["LOGURU_INFO_COLOR"] = "<bold><green>"
    from loguru import logger  # type: ignore

    if not logger._core.handlers:  # type: ignore

        def log_filter(record):
            if record["function"] == "<module>":
                record["function"] = "\b"
            return True

        log_format = (
            "<level>{level: <8}</level> <bold><level>|</level></bold> "
            "{time:YYYY-MM-DD HH:mm:ss} <bold><level>|</level></bold> "
            "{name}:{function}:{line}<bold><level> > </level></bold><level>{message}</level>"
        )
        logger.add(sys.stderr, format=log_format, enqueue=True, colorize=True, filter=log_filter)
        warnings.showwarning = showwarning

        class InterceptHandler(logging.Handler):
            def emit(self, record):
                try:
                    level = logger.level(record.levelname).name  # type: ignore
                except ValueError:
                    level = record.levelno
                frame, depth = sys._getframe(6), 6
                while frame and frame.f_code.co_filename == logging.__file__:
                    frame = frame.f_back  # type: ignore
                    depth += 1

                logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())  # type: ignore

        logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)


elif logging_module == "logging":
    """
    This is workaround for lazy formating with {} in logging.

    About:
    https://docs.python.org/3/howto/logging-cookbook.html#use-of-alternative-formatting-styles
    """
    import logging

    import colorama

    colorama.just_fix_windows_console()
    LEVEL_COLORS = {
        "DEBUG": colorama.Style.BRIGHT + colorama.Fore.BLUE,
        "INFO": colorama.Style.BRIGHT + colorama.Fore.GREEN,
        "WARNING": colorama.Fore.YELLOW,
        "ERROR": colorama.Fore.RED,
        "CRITICAL": colorama.Style.BRIGHT + colorama.Fore.RED,
    }

    loguru_like_format = (
        "<level>{levelname: <8}</level> <bold><level>|</level></bold> "
        "{asctime} <bold><level>|</level></bold> "
        "{module}:{funcName}:{lineno}<bold><level> > </level></bold><level>{message}</level>"
    )

    class ColorFormatter(logging.Formatter):
        def format(self, record):
            color = LEVEL_COLORS.get(record.levelname, "")
            log_format = (
                loguru_like_format.replace("<level>", color)
                .replace("</level>", colorama.Style.RESET_ALL)
                .replace("<bold>", colorama.Style.BRIGHT)
                .replace("</bold>", colorama.Style.RESET_ALL)
            )
            if not record.funcName or record.funcName == "<module>":
                record.funcName = "\b"
            frame = next(
                (
                    frame
                    for frame in inspect.stack()
                    if frame.filename == record.pathname and frame.lineno == record.lineno
                ),
                None,
            )
            if frame:
                module = inspect.getmodule(frame.frame)
                record.module = module.__name__ if module else "<module>"
            return logging.Formatter(
                log_format,
                datefmt=self.datefmt,
                style="{",
            ).format(record)

    logging.basicConfig(level=logging.DEBUG)
    logging.root.handlers[0].setFormatter(ColorFormatter())

    class LogMessage:
        def __init__(self, fmt, args, kwargs):
            self.fmt = fmt
            self.args = args
            self.kwargs = kwargs

        def __str__(self):
            return self.fmt.format(*self.args)

    class StyleAdapter(logging.LoggerAdapter):
        def __init__(self, logger, extra=None):
            super().__init__(logger, extra or {})

        def log(self, level, msg, *args, **kwargs):
            if self.isEnabledFor(level):
                if "stacklevel" not in kwargs:
                    kwargs["stacklevel"] = 2
                msg, args, kwargs = self.process(msg, args, kwargs)
                self.logger._log(level, msg, args, **kwargs)

        def process(self, msg, args, kwargs):  # type: ignore
            log_kwargs = {
                key: kwargs[key]
                for key in inspect.getfullargspec(self.logger._log).args[1:]
                if key in kwargs
            }
            if isinstance(msg, str):
                msg = LogMessage(msg, args, kwargs)
                args = ()
            return msg, args, log_kwargs

    warnings.showwarning = showwarning

    logger = StyleAdapter(logging.getLogger("vkbottle"))  # type: ignore
    logger.info("logging is used as the default logger, but we recommend using loguru instead")

if hasattr(asyncio, "WindowsProactorEventLoopPolicy") and isinstance(
    asyncio.get_event_loop_policy(),
    asyncio.WindowsProactorEventLoopPolicy,  # type: ignore
):
    """
    This is a workaround for a bug in ProactorEventLoop:
    https://github.com/aio-libs/aiohttp/issues/4324

    This also can be fixed by using loop.run_until_complete instead of asyncio.run
    but I like to use asyncio.run because it's more readable, and not require to create new event loop.
    """
    from asyncio.proactor_events import _ProactorBasePipeTransport, _ProactorBaseWritePipeTransport
    from functools import wraps

    def silence_exception(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except (AttributeError, RuntimeError) as e:
                if str(e) not in (
                    "'NoneType' object has no attribute 'send'",
                    "Event loop is closed",
                ):
                    raise

        return wrapper

    _ProactorBasePipeTransport.__del__ = silence_exception(_ProactorBasePipeTransport.__del__)  # type: ignore
    _ProactorBaseWritePipeTransport._loop_writing = silence_exception(  # type: ignore
        _ProactorBaseWritePipeTransport._loop_writing  # type: ignore
    )



from typing import Optional

from vkbottle.modules import pydantic


class ShowSnackbarEvent(pydantic.BaseModel):
    type: str = pydantic.Field(default="show_snackbar", const=True)
    text: str


class OpenLinkEvent(pydantic.BaseModel):
    type: str = pydantic.Field(default="open_link", const=True)
    link: str


class OpenAppEvent(pydantic.BaseModel):
    type: str = pydantic.Field(default="open_app", const=True)
    owner_id: Optional[int] = None
    app_id: int
    hash: str


__all__ = (
    "OpenAppEvent",
    "OpenLinkEvent",
    "ShowSnackbarEvent",
)



from typing import Any, Hashable, Optional

from vkbottle.tools.ctx_tool import BaseContext

from .abc import ABCStorage


class CtxStorage(ABCStorage, BaseContext):
    """Context storage
    Documentation: https://vkbottle.rtfd.io/ru/latest/tools/storage
    """

    storage: dict

    def __init__(
        self,
        default: Optional[dict] = None,
        force_reset: bool = False,
    ):
        if not self.get_instance() or force_reset:
            default = default or {}
            self.storage = default
            self.set_instance(self)

    def set(self, key: Hashable, value: Any) -> None:
        current_storage = self.get_instance().storage
        current_storage[key] = value
        self.set_instance(CtxStorage(current_storage, True))

    def get(self, key: Hashable) -> Any:
        return self.get_instance().storage.get(key)

    def delete(self, key: Hashable) -> None:
        new_storage = self.get_instance().storage
        new_storage.pop(key)
        self.set_instance(CtxStorage(new_storage, True))

    def contains(self, key: Hashable) -> bool:
        return key in self.get_instance().storage



from .audio import AudioUploader
from .base import BaseUploader
from .doc import (
    DocMessagesUploader,
    DocUploader,
    DocWallUploader,
    GraffitiUploader,
    VoiceMessageUploader,
)
from .photo import (
    PhotoChatFaviconUploader,
    PhotoFaviconUploader,
    PhotoMarketUploader,
    PhotoMessageUploader,
    PhotoToAlbumUploader,
    PhotoUploader,
    PhotoWallUploader,
)
from .video import VideoUploader

__all__ = (
    "AudioUploader",
    "BaseUploader",
    "DocMessagesUploader",
    "DocUploader",
    "DocWallUploader",
    "GraffitiUploader",
    "PhotoChatFaviconUploader",
    "PhotoFaviconUploader",
    "PhotoMarketUploader",
    "PhotoMessageUploader",
    "PhotoToAlbumUploader",
    "PhotoUploader",
    "PhotoWallUploader",
    "VideoUploader",
    "VoiceMessageUploader",
)



import asyncio
import contextlib
import warnings  # type: ignore
from asyncio import new_event_loop
from typing import TYPE_CHECKING, Any, Callable, Coroutine, List, Optional, Union

from typing_extensions import deprecated  # type: ignore

from vkbottle.modules import logger

from .delayed_task import DelayedTask

if TYPE_CHECKING:
    Task = Coroutine[Any, Any, Any]
    from asyncio import AbstractEventLoop


class LoopWrapper:
    """Loop Wrapper for vkbottle manages startup, shutdown and main tasks,
    creates loop and runs it forever"""

    def __init__(
        self,
        *,
        on_startup: Optional[List["Task"]] = None,
        on_shutdown: Optional[List["Task"]] = None,
        tasks: Optional[List["Task"]] = None,
        loop: Optional["AbstractEventLoop"] = None,
    ) -> None:
        self.on_startup = on_startup or []
        self.on_shutdown = on_shutdown or []
        self.tasks = tasks or []
        self.loop = loop

    @property
    @deprecated(
        "LoopWrapper.auto_reload is deprecated, instead, install watchfiles",
        stacklevel=0,
    )
    def auto_reload(self) -> bool:
        return False

    @auto_reload.setter
    def auto_reload(self, value: bool) -> None:
        warnings.warn(
            "LoopWrapper.auto_reload is deprecated, instead, install watchfiles",
            DeprecationWarning,
            stacklevel=0,
        )

    @deprecated(
        "Deprecated. Use run() instead",
        stacklevel=0,
    )
    def run_forever(self):
        logger.warning("run_forever is deprecated. Use run() instead")
        self.run()

    def run(self) -> None:
        """Runs startup tasks and makes the loop running until all tasks are done"""

        if not self.tasks:
            logger.warning("You ran loop with 0 tasks. Is it ok?")

        self.loop = new_event_loop() if self.loop is None else self.loop

        for startup_task in self.on_startup:
            self.loop.run_until_complete(startup_task)

        while self.tasks:
            self.loop.create_task(self.tasks.pop(0))

        tasks = asyncio.all_tasks(self.loop)
        try:
            while tasks:
                tasks_results, _ = self.loop.run_until_complete(
                    asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
                )
                for task_result in tasks_results:
                    try:
                        task_result.result()
                    except Exception as exc:  # noqa: PERF203
                        logger.exception(exc)
                tasks = asyncio.all_tasks(self.loop)
        except KeyboardInterrupt:
            logger.info("Caught keyboard interrupt. Shutting down...")
            task_to_cancel = asyncio.gather(*tasks)
            task_to_cancel.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                self.loop.run_until_complete(task_to_cancel)
        finally:
            for shutdown_task in self.on_shutdown:
                self.loop.run_until_complete(shutdown_task)
            if self.loop.is_running():
                self.loop.close()

    def add_task(self, task: Union["Task", Callable[..., "Task"]]):
        """Adds tasks to be ran in run_forever or run it immediately if loop is already running
        :param task: coroutine / coroutine function with zero arguments
        """
        if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):
            task = task()  # type: ignore
        elif not asyncio.iscoroutine(task):
            msg = "Task should be coroutine or coroutine function"
            raise TypeError(msg)

        if self.loop and self.loop.is_running():
            self.loop.create_task(task)
        else:
            self.tasks.append(task)

    def interval(
        self,
        seconds: int = 0,
        minutes: int = 0,
        hours: int = 0,
        days: int = 0,
    ) -> Callable[[Callable], Callable]:
        """A tiny template to wrap repeated tasks with decorator
        >>> lw = LoopWrapper()
        >>> @lw.interval(seconds=5)
        >>> async def repeated_function():
        >>>     print("This will be logged every five seconds")
        >>> lw.run()
        """

        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: Callable):
            self.add_task(DelayedTask(seconds, func))
            return func

        return decorator

    def timer(
        self,
        seconds: int = 0,
        minutes: int = 0,
        hours: int = 0,
        days: int = 0,
    ) -> Callable[[Callable], Callable]:
        """A tiny template to wrap tasks with timer
        >>> lw = LoopWrapper()
        >>> @lw.timer(seconds=5)
        >>> async def delayed_function():
        >>>     print("This will after 5 seconds")
        >>> lw.run()
        """
        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: Callable):
            self.add_task(DelayedTask(seconds, func, do_break=True))
            return func

        return decorator


__all__ = ("LoopWrapper",)



from __future__ import annotations

import typing
from collections import UserDict, deque

KT = typing.TypeVar("KT")
VT = typing.TypeVar("VT")

if typing.TYPE_CHECKING:

    class LimitedDict(UserDict[KT, VT]):
        maxlimit: int
        queue: deque[KT]

        def __init__(self, *, maxlimit: int = 1000) -> None: ...

        def set(self, key: KT, value: VT, /) -> typing.Optional[VT]:
            """Set item in the dictionary.
            Returns a value that was deleted when the limit in the dictionary
            was reached, otherwise None.
            """

        def __setitem__(self, key: KT, value: VT, /) -> None: ...

        def __delitem__(self, key: KT) -> None: ...

else:

    class LimitedDict(UserDict, typing.Generic[KT, VT]):
        def __init__(self, *, maxlimit=1000):
            super().__init__()
            self.maxlimit = maxlimit
            self.queue = deque(maxlen=maxlimit)

        def set(self, key, value, /):
            deleted_item = None
            if len(self.queue) >= self.maxlimit:
                deleted_item = self.pop(self.queue.popleft(), None)
            if key not in self.queue:
                self.queue.append(key)
            super().__setitem__(key, value)
            return deleted_item

        def __setitem__(self, key, value, /) -> None:
            self.set(key, value)

        def __delitem__(self, key) -> None:
            if key in self.queue:
                self.queue.remove(key)
            return super().__delitem__(key)


__all__ = ("LimitedDict",)



from random import choice
from string import ascii_lowercase
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from typing_extensions import Self

from vkbottle.exception_factory import ErrorHandler
from vkbottle.modules import logger

from .abc import ABCCallback

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.exception_factory import ABCErrorHandler


class BotCallback(ABCCallback):
    """Bot Callback class"""

    def __init__(
        self,
        url: Optional[str] = None,
        title: Optional[str] = None,
        secret_key: Optional[str] = None,
        api: Optional["ABCAPI"] = None,
        group_id: Optional[int] = None,
        error_handler: Optional["ABCErrorHandler"] = None,
    ) -> None:
        self.url = url
        self.title = title
        self.secret_key = secret_key or self._generate_secret_key()
        self._api = api
        self.group_id = group_id
        self.error_handler = error_handler or ErrorHandler()  # type: ignore

    def get_secret_key(self) -> str:
        return self.secret_key

    def _generate_secret_key(self) -> str:
        return "".join(choice(ascii_lowercase) for _ in range(32))

    async def setup_group_id(self):
        if self.group_id is None:
            self.group_id = (await self.api.request("groups.getById", {}))["response"][0]["id"]

    async def find_server_id(self) -> Optional[int]:
        servers = await self.get_callback_servers()
        if servers:
            return next((server["id"] for server in servers if server["url"] == self.url), None)
        return None

    async def add_callback_server(self) -> int:
        logger.debug("Adding callback server...")
        data = {
            "group_id": self.group_id,
            "url": self.url,
            "title": self.title,
            "secret_key": self.secret_key,
        }
        return (await self.api.request("groups.addCallbackServer", data))["response"]["server_id"]

    async def delete_callback_server(self, server_id: int) -> None:
        logger.debug("Delete callback server...")
        await self.api.request(
            "groups.deleteCallbackServer", {"group_id": self.group_id, "server_id": server_id}
        )

    async def edit_callback_server(self, server_id: int, secret_key: Optional[str] = None) -> None:
        logger.debug("Editing callback server...")
        data = {
            "group_id": self.group_id,
            "server_id": server_id,
            "url": self.url,
            "title": self.title,
            "secret_key": self.secret_key,
        }

        if secret_key is not None:
            data["secret_key"] = secret_key
        await self.api.request("groups.editCallbackServer", data)

    async def get_callback_confirmation_code(self) -> str:
        logger.debug("Getting callback confirmation code...")
        return (
            await self.api.request(
                "groups.getCallbackConfirmationCode",
                {
                    "group_id": self.group_id,
                },
            )
        )["response"]["code"]

    async def get_callback_servers(
        self,
        servers_ids: Optional[List[int]] = None,
    ) -> List[Dict[str, Any]]:
        logger.debug("Getting callback servers...")
        data: Dict[str, Any] = {"group_id": self.group_id}
        if servers_ids is not None:
            data["server_ids"] = ",".join(map(str, servers_ids))
        return (await self.api.request("groups.getCallbackServers", data))["response"]["items"]

    async def get_callback_settings(self, server_id: int) -> Dict[str, bool]:
        logger.debug("Getting callback settings...")
        return (
            await self.api.request(
                "groups.getCallbackSettings", {"group_id": self.group_id, "server_id": server_id}
            )
        )["response"]["events"]

    async def set_callback_settings(
        self,
        server_id: int,
        params: Optional[Dict[str, bool]] = None,
    ) -> None:
        """Search values in https://dev.vk.com/method/groups.getCallbackSettings"""

        logger.debug("Setting callback settings...")

        data = {"group_id": self.group_id, "server_id": server_id}

        if params is not None:
            for k, v in params.items():
                data[k] = v  # noqa: PERF403
        await self.api.request("groups.setCallbackSettings", data)

    def construct(
        self,
        api: "ABCAPI",
        error_handler: Optional["ABCErrorHandler"] = None,
    ) -> Self:
        self._api = api
        if error_handler is not None:
            self.error_handler = error_handler
        return self

    @property
    def api(self) -> "ABCAPI":
        if self._api is None:
            msg = "You must construct callback with API before try to access api property of Callback"
            raise NotImplementedError(msg)
        return self._api

    @api.setter
    def api(self, new_api: "ABCAPI") -> None:
        self._api = new_api



import asyncio
import importlib
import os
import re
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Coroutine, Iterator, TypeVar

from vkbottle.modules import logger

if TYPE_CHECKING:
    from vkbottle.framework.abc_blueprint import ABCBlueprint

T = TypeVar("T")


# This feature is not used in production
# but can be useful for customization
# purposes
def run_in_task(coroutine: Coroutine[Any, Any, Any]) -> asyncio.Task:
    """Gets loop and runs add makes task from the given coroutine"""

    loop = asyncio.get_running_loop()
    return loop.create_task(coroutine)


def run_sync(coroutine: Coroutine[Any, Any, T]) -> T:
    return ThreadPoolExecutor().submit(asyncio.run, coroutine).result()  # type: ignore


def load_blueprints_from_package(package_name: str) -> Iterator["ABCBlueprint"]:
    """Gets blueprints from package
    >>> for bp in load_blueprints_from_package("blueprints"):
    >>>     bp.load(...)
    """

    logger.warning(
        "Blueprints was deprecated and will be removed in future releases, read about new code separation method in documentation: \n"
        "https://vkbottle.readthedocs.io/ru/latest/tutorial/code-separation/"
    )
    bp_paths = []
    for filename in os.listdir(package_name):
        if filename.startswith("__"):
            continue
        elif not filename.endswith(".py"):
            yield from load_blueprints_from_package(os.path.join(package_name, filename))
            continue

        with open(os.path.join(package_name, filename), encoding="utf-8") as file:
            bp_names = re.findall(
                r"^(\w+) = (?:Bot|User|)Blueprint\(", file.read(), flags=re.MULTILINE
            )

            if len(bp_names) != 1:
                msg = f"Invalid blueprint declaration in file: {filename}"
                raise ValueError(msg)
            bp_paths.append((filename[:-3], bp_names[0]))

    for bp_path in bp_paths:
        module, bp_name = bp_path
        module_name = package_name.replace(f".{os.sep}", ".").replace(os.sep, ".")
        bp_module = importlib.import_module(f"{module_name}.{module}")
        yield getattr(bp_module, bp_name)



import sys
import warnings

warnings.warn(
    "Imports from vkbottle.tools.dev is deprecated, use vkbottle.tools instead",
    DeprecationWarning,
    stacklevel=0,
)

sys.modules["vkbottle.tools.dev"] = sys.modules["vkbottle.tools"]



import datetime
import typing

from vkbottle.dispatch.handlers.from_func_handler import FromFuncHandler
from vkbottle.dispatch.middlewares.abc import BaseMiddleware

if typing.TYPE_CHECKING:
    from .machine import WaiterMachine
    from .short_state import ShortState


class WaiterMiddleware(BaseMiddleware[dict]):
    def __init__(self, machine: "WaiterMachine") -> None:
        self.machine = machine

    def __call__(self, *args: typing.Any, **kw: typing.Any) -> "WaiterMiddleware":
        middleware = self.__class__(self.machine)
        super(WaiterMiddleware, middleware).__init__(*args, **kw)
        return middleware

    async def pre(self) -> None:
        if not self.view or not hasattr(self.view, "get_state_key"):
            msg = "WaiterMiddleware cannot be used inside a view which doesn't provide get_state_key (ABCDispenseView interface)"
            raise RuntimeError(msg)

        view_name = self.view.__class__.__name__
        if view_name not in self.machine.storage:
            return

        key = self.view.get_state_key(self.event)
        short_state: typing.Optional["ShortState"] = self.machine.storage[view_name].get(key)
        if not short_state:
            return

        if (
            short_state.expiration_date is not None
            and datetime.datetime.now() >= short_state.expiration_date
        ):
            await self.machine.drop(self.view, short_state.key)  # type: ignore
            return

        context = {}
        if short_state.context is not None:
            context.update(short_state.context.context)

        if short_state.exit_behaviour is not None and await self.machine.call_behaviour(
            self.view,  # type: ignore
            short_state.exit_behaviour,
            self.event,
            **context,
        ):
            await self.machine.drop(self.view, short_state.key)  # type: ignore
            return

        handler: FromFuncHandler = FromFuncHandler(self.pass_runtime, *short_state.rules)
        result = await handler.filter(self.event)

        if result is not False:
            if isinstance(result, bool):
                result = {}
            result.update(context)
            await handler.handle(self.event, short_state=short_state, **result)

        elif short_state.default_behaviour is not None:
            await self.machine.call_behaviour(
                self.view,  # type: ignore
                short_state.default_behaviour,
                self.event,
                **context,
            )

        self.stop("Runtime was passed to waiter.")

    async def pass_runtime(
        self, event: typing.Any, short_state: "ShortState[dict]", **context: typing.Any
    ) -> None:
        short_state.context = (event, context)  # type: ignore
        short_state.event.set()



import contextlib
from typing import TYPE_CHECKING, Any, Optional, Union

from vkbottle.modules import json, logger

from .abc import ABCResponseValidator

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


class JSONResponseValidator(ABCResponseValidator):
    """Default response json-parse validator
    Documentation: https://vkbottle.rtfd.io/ru/latest/low-level/api/response-validator
    """

    def __init__(self, context: Optional[dict] = None):
        self.context = context or {}

    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Any:
        if isinstance(response, dict):
            return response
        elif isinstance(response, str):
            with contextlib.suppress(ValueError):
                return json.loads(response)

        if self.context.get("reschedule"):
            return None

        logger.info(
            "VK returned object of invalid type ({}). Request will be rescheduled with {}",
            type(response).__name__,
            ctx_api.request_rescheduler.__class__.__name__,
        )
        self.context["reschedule"] = True
        response = await self.validate(
            method,
            data,
            await ctx_api.request_rescheduler.reschedule(ctx_api, method, data, response),
            ctx_api,
        )
        self.context.pop("reschedule")
        return response



from typing import TYPE_CHECKING

from vkbottle.modules import json

if TYPE_CHECKING:
    from .element import TemplateElement


def template_gen(*e: "TemplateElement"):
    """Generates dumped carousel out of box"""
    return json.dumps({"type": "carousel", "elements": [e.raw for e in e]})



import os.path
from typing import TYPE_CHECKING, Optional, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class DocUploader(BaseUploader):
    NAME = "doc.txt"

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getUploadServer", kwargs))["response"]

    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> str:
        doc = await self.raw_upload(file_source, group_id, **params)
        doc_type = doc["type"]
        return self.generate_attachment_string(
            doc_type,
            doc[doc_type]["owner_id"],
            doc[doc_type]["id"],
            doc[doc_type].get("access_key"),
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> dict:
        title = params.pop("title", None)
        if title is None:
            title = os.path.split(file_source)[1] if isinstance(file_source, str) else self.NAME
        server = await self.get_server(group_id=group_id, **params)
        data = await self.read(file_source)

        file = self.get_bytes_io(data, name=title)

        uploader = await self.upload_files(server["upload_url"], {"file": file})

        return (
            await self.api.request(
                "docs.save",
                {"title": title, **uploader, **params},
            )
        )["response"]

    @property
    def attachment_name(self) -> str:
        return self._attachment_name or self.NAME


class DocWallUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getWallUploadServer", kwargs))["response"]


class DocMessagesUploader(DocUploader):
    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        peer_id: Optional[int] = None,
        **params,
    ) -> str:
        return await super().upload(
            file_source=file_source,
            group_id=group_id,
            peer_id=peer_id,
            **params,
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        peer_id: Optional[int] = None,
        **params,
    ) -> dict:
        return await super().raw_upload(
            file_source=file_source,
            group_id=group_id,
            peer_id=peer_id,
            **params,
        )

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getMessagesUploadServer", kwargs))["response"]


class VoiceMessageUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (
            await self.api.request(
                "docs.getMessagesUploadServer", {"type": "audio_message", **kwargs}
            )
        )["response"]


class GraffitiUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (
            await self.api.request("docs.getMessagesUploadServer", {"type": "graffiti", **kwargs})
        )["response"]


__all__ = (
    "DocMessagesUploader",
    "DocUploader",
    "DocWallUploader",
    "GraffitiUploader",
    "VoiceMessageUploader",
)



from enum import Enum
from typing import Any, Dict, cast

from vkbottle.modules import pydantic


class BaseStateGroup(str, Enum):
    __slots__ = ()

    def __str__(self) -> str:
        return get_state_repr(self)


def get_state_repr(state: BaseStateGroup) -> str:
    return f"{state.__class__.__name__}:{state.value}"


class StateRepresentation(str):
    __slots__ = ()

    def __eq__(self, __x: object) -> bool:
        if isinstance(__x, BaseStateGroup):
            return self == get_state_repr(__x)
        return super().__eq__(__x)


class StatePeer(pydantic.BaseModel):
    peer_id: int
    state: str
    payload: Dict[str, Any] = cast(Dict[str, Any], pydantic.Field(default_factory=dict))

    @pydantic.validator("state", pre=True)
    def validate_state(cls, v: Any) -> str:
        if isinstance(v, BaseStateGroup):
            return StateRepresentation(v)
        elif isinstance(v, str):
            return v
        msg = f"State value must be `string` or `BaseStateGroup`, got `{type(v)}`"
        raise ValueError(msg)



from typing import TYPE_CHECKING, Optional, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class VideoUploader(BaseUploader):
    NAME = "video.mp4"

    @property
    def attachment_name(self) -> str:
        return self._attachment_name or self.NAME

    async def upload(
        self,
        file_source: Optional[Union[str, "Bytes"]] = None,
        link: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_private: Optional[bool] = None,
        wallpost: Optional[bool] = None,
        group_id: Optional[int] = None,
        album_id: Optional[int] = None,
        privacy_view: Optional[list] = None,
        privacy_comment: Optional[list] = None,
        no_comments: Optional[bool] = None,
        repeat: Optional[bool] = None,
        compression: Optional[bool] = None,
        **params,
    ) -> str:
        server = await self.raw_upload(
            file_source=file_source,
            link=link,
            name=name,
            description=description,
            is_private=is_private,
            wallpost=wallpost,
            group_id=group_id,
            album_id=album_id,
            privacy_view=privacy_view,
            privacy_comment=privacy_comment,
            no_comments=no_comments,
            repeat=repeat,
            compression=compression,
            **params,
        )
        return self.generate_attachment_string(
            "video",
            server["owner_id"],
            server["video_id"],
        )

    async def raw_upload(
        self,
        file_source: Optional[Union[str, "Bytes"]] = None,
        link: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_private: Optional[bool] = None,
        wallpost: Optional[bool] = None,
        group_id: Optional[int] = None,
        album_id: Optional[int] = None,
        privacy_view: Optional[list] = None,
        privacy_comment: Optional[list] = None,
        no_comments: Optional[bool] = None,
        repeat: Optional[bool] = None,
        compression: Optional[bool] = None,
        **params,
    ) -> dict:
        server = await self.get_server(
            group_id=group_id,
            album_id=album_id,
            name=name,
            description=description,
            is_private=is_private,
            wallpost=wallpost,
            link=link,
            privacy_view=privacy_view,
            privacy_comment=privacy_comment,
            no_comments=no_comments,
            repeat=repeat,
            compression=compression,
        )
        if file_source is None and link is None:
            msg = "You must specify either file_source or link"
            raise ValueError(msg)

        if link:
            await self.api.http_client.request_json(server["upload_url"], method="GET")
        else:
            data = await self.read(file_source)  # type: ignore
            file = self.get_bytes_io(data)
            await self.upload_files(server["upload_url"], {"video_file": file}, **params)

        return server

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("video.save", kwargs))["response"]


__all__ = ("VideoUploader",)



import dataclasses
from typing import TYPE_CHECKING, Any, Dict, List, Type

from vkbottle_types.events import UserEventType

from vkbottle.dispatch.return_manager.user import UserMessageReturnHandler
from vkbottle.dispatch.views.abc.raw import ABCRawEventView, BaseHandlerBasement

if TYPE_CHECKING:
    from vkbottle_types.events.user_events import BaseUserEvent

    from vkbottle.dispatch.handlers import ABCHandler


@dataclasses.dataclass(frozen=True)
class UserHandlerBasement(BaseHandlerBasement):
    dataclass: Type["BaseUserEvent"]
    handler: "ABCHandler"


class RawUserEventView(ABCRawEventView[List[Any], UserHandlerBasement]):
    handlers: Dict[UserEventType, List[UserHandlerBasement]]

    def __init__(self) -> None:
        super().__init__()
        self.handlers = {}
        self.handler_return_manager = UserMessageReturnHandler()

    def get_handler_basements(self, event: List[Any]) -> List[UserHandlerBasement]:
        return self.handlers[UserEventType(self.get_event_type(event))]

    def get_event_model(
        self,
        handler_basement: UserHandlerBasement,
        event: List[Any],
    ) -> "BaseUserEvent":
        return handler_basement.dataclass.parse(event[1:])

    @staticmethod
    def get_event_type(event: List[Any]) -> int:
        return event[0]

    async def process_event(self, event: List[Any]) -> bool:
        try:
            event_type = UserEventType(self.get_event_type(event))
        except ValueError:
            event_type = UserEventType.UNDEFINED_EVENT

        return event_type in self.handlers



from abc import abstractmethod
from typing import TYPE_CHECKING, Any, NoReturn, Optional, Union

from typing_extensions import deprecated  # type: ignore

from .abc import ABCFramework

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API
    from vkbottle.dispatch import ABCRouter, ABCStateDispenser
    from vkbottle.polling import ABCPolling

CONSTRUCT_BLUEPRINT = "You need to construct blueprint firstly"


@deprecated(
    "Blueprints was deprecated and will be removed in future releases, "
    "read about new code separation method in documentation: \n"
    "https://vkbottle.rtfd.io/ru/latest/tutorial/code-separation/",
    stacklevel=0,
)
class ABCBlueprint(ABCFramework):
    router: "ABCRouter"

    _polling: "ABCPolling"
    _api: "ABCAPI"
    _state_dispenser: "ABCStateDispenser"

    name: str = "Unnamed"
    constructed: bool = False

    @abstractmethod
    def construct(
        self, api: "ABCAPI", polling: "ABCPolling", state_dispenser: "ABCStateDispenser"
    ) -> "ABCBlueprint":
        pass

    @abstractmethod
    def load(self, framework: Any) -> "ABCBlueprint":
        pass

    @property
    def polling(self) -> "ABCPolling":
        self.assert_constructed()
        return self._polling

    @polling.setter
    def polling(self, new_polling: "ABCPolling"):  # type: ignore
        self._polling = new_polling

    @property
    def state_dispenser(self) -> "ABCStateDispenser":
        self.assert_constructed()
        return self._state_dispenser

    @state_dispenser.setter
    def state_dispenser(self, new_state_dispenser: "ABCStateDispenser"):
        self._state_dispenser = new_state_dispenser

    @property  # type: ignore
    def api(self) -> Union["ABCAPI", "API"]:  # type: ignore
        if not self._api:
            raise RuntimeError(
                CONSTRUCT_BLUEPRINT
                + " Beware: if you use multibot, api can only be accessed with event.ctx_api"
            )
        return self._api

    @api.setter
    def api(self, new_api: "ABCAPI"):
        self._api = new_api

    async def run_polling(self) -> NoReturn:
        msg = "You are not allowed to run polling with blueprint"
        raise RuntimeError(msg)

    def run_forever(self) -> NoReturn:
        msg = "You are not allowed to run polling with blueprint"
        raise RuntimeError(msg)

    def assert_constructed(self) -> Optional[NoReturn]:
        if not self.constructed:
            raise RuntimeError(CONSTRUCT_BLUEPRINT)

    def __repr__(self):
        return f"<Blueprint {self.name!r} {self.__class__.__qualname__} constructed={self.constructed}>"



import typing

from .from_func_handler import FromFuncHandler

if typing.TYPE_CHECKING:
    from vkbottle.dispatch.rules import ABCRule
    from vkbottle.tools.mini_types.base import BaseMessageMin


class MessageReplyHandler(FromFuncHandler["BaseMessageMin"]):
    def __init__(
        self,
        text: str,
        *rules: "ABCRule[BaseMessageMin]",
        is_blocking: bool = True,
        as_reply: bool = False,
        **default_params: typing.Any,
    ) -> None:
        super().__init__(self.reply_handler, *rules, blocking=is_blocking)
        self.text = text
        self.as_reply = as_reply
        self.default_params = default_params

    async def reply_handler(self, message: "BaseMessageMin") -> None:
        params = {"message": self.text, "random_id": 0, **self.default_params}
        method = message.reply if self.as_reply else message.answer
        await method(**params)  # type: ignore



import warnings
from abc import ABC, abstractmethod
from io import BytesIO
from typing import TYPE_CHECKING, Optional, Union

import aiofiles

from vkbottle.exception_factory.base_exceptions import VKAPIError
from vkbottle.modules import json

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI

    Bytes = Union[bytes, BytesIO]


class BaseUploader(ABC):
    def __init__(
        self,
        api: "ABCAPI",
        attachment_name: Optional[str] = None,
        **kwargs,
    ):
        self.api = api
        if "generate_attachment_strings" in kwargs:
            warnings.warn(
                "generate_attachment_strings in uploaders is deprecated"
                " use .raw_upload() to get raw response or .upload() to get attachment string",
                DeprecationWarning,
                stacklevel=0,
            )
            kwargs.pop("generate_attachment_strings")
        self._attachment_name = attachment_name

    @abstractmethod
    async def get_server(self, **kwargs) -> dict:
        pass

    @property
    @abstractmethod
    def attachment_name(self) -> str:
        pass

    async def upload_files(self, upload_url: str, files: dict) -> dict:
        raw_response = await self.api.http_client.request_text(
            upload_url, method="POST", data=files
        )
        return json.loads(raw_response)

    def get_bytes_io(self, data: "Bytes", name: Optional[str] = None) -> BytesIO:
        bytes_io = data if isinstance(data, BytesIO) else BytesIO(data)
        # To avoid errors with image generators (such as pillow)
        bytes_io.seek(0)
        # To guarantee VK API file extension recognition
        if not hasattr(bytes_io, "name"):
            bytes_io.name = name or self.attachment_name
        return bytes_io

    async def get_owner_id(self, **upload_params) -> int:
        if "group_id" in upload_params:
            return upload_params["group_id"]
        if "user_id" in upload_params:
            return upload_params["user_id"]
        if "owner_id" in upload_params:
            return upload_params["owner_id"]
        try:
            return -(await self.api.request("groups.getById", {}))["response"][0]["id"]
        except VKAPIError:
            return (await self.api.request("users.get", {}))["response"][0]["id"]

    @staticmethod
    def generate_attachment_string(
        attachment_type: str, owner_id: int, item_id: int, access_key: Optional[str] = None
    ) -> str:
        return f"{attachment_type}{owner_id}_{item_id}{f'_{access_key}' if access_key else ''}"

    @staticmethod
    async def read(file_source: Union[str, "Bytes"]) -> "Bytes":
        if isinstance(file_source, str):
            async with aiofiles.open(file_source, "rb") as file:
                return await file.read()
        return file_source

    def __repr__(self) -> str:
        return f"<Uploader {self.__class__.__name__} with api {self.api!r}"



from .abc import ABCTokenGenerator


class SingleTokenGenerator(ABCTokenGenerator):
    def __init__(self, token: str):
        self.token = token

    async def get_token(self) -> str:
        return self.token

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



from typing import TYPE_CHECKING, Any, Union

from vkbottle.exception_factory import CaptchaError, VKAPIError
from vkbottle.modules import logger

from .abc import ABCResponseValidator

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


class VKAPIErrorResponseValidator(ABCResponseValidator):
    """Default vk api error response validator
    Documentation: https://vkbottle.rtfd.io/ru/latest/low-level/api/response-validator
    """

    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Any:
        if "error" not in response:
            if "response" not in response:
                request_params = [{"key": key, "value": value} for key, value in data.items()]
                raise VKAPIError[1](
                    error_msg=f"Unknown response from {method}: {response}",
                    request_params=request_params,
                )
            if isinstance(response["response"], list) and any(
                "error" in item for item in response["response"]
            ):
                logger.info("API error(s) in response wasn't handled: {}", response["response"])
            return response

        if ctx_api.ignore_errors:
            return None
        error = response["error"]
        code = error.pop("error_code")

        if VKAPIError[code] is CaptchaError and ctx_api.captcha_handler:
            key = await ctx_api.captcha_handler(CaptchaError(**error))  # type: ignore
            return await ctx_api.request(
                method, {**data, "captcha_sid": error["captcha_sid"], "captcha_key": key}
            )

        raise VKAPIError[code](**error)



import typing
from abc import ABC, abstractmethod
from typing import Any, Union

Event = typing.TypeVar("Event")


class ABCHandler(ABC, typing.Generic[Event]):
    blocking: bool

    @abstractmethod
    async def filter(self, event: Event) -> Union[dict, bool]:
        pass

    @abstractmethod
    async def handle(self, event: Event, **context) -> Any:
        pass



from .base_exceptions import APIAuthError, CaptchaError, VKAPIError
from .code_exception import CodeException
from .error_handler import ABCErrorHandler, ErrorHandler
from .reducible_kwargs_exception import ReducibleKwargsException

__all__ = (
    "ABCErrorHandler",
    "CaptchaError",
    "APIAuthError",
    "CodeException",
    "ErrorHandler",
    "VKAPIError",
    "ReducibleKwargsException",
)



from typing import TYPE_CHECKING, Any, Callable, List, Optional, Type, Union

from vkbottle_types.events import GroupEventType

from vkbottle.dispatch.handlers import FromFuncHandler
from vkbottle.dispatch.rules import ABCRule
from vkbottle.dispatch.views.bot import BotHandlerBasement, BotMessageView, RawBotEventView

from .base import BaseLabeler, CustomRuleType

if TYPE_CHECKING:
    from vkbottle_types.events import BaseGroupEvent

    from vkbottle.dispatch.views.bot import ABCBotMessageView
    from vkbottle.tools.mini_types.bot.message import MessageMin

    from .abc import LabeledHandler

    LabeledMessageHandler = Callable[..., Callable[["MessageMin"], Any]]
    EventName = Union[GroupEventType, str]


class BotLabeler(BaseLabeler):
    """BotLabeler - shortcut manager for router
    Can be loaded to other BotLabeler
    >>> bl = BotLabeler()
    >>> ...
    >>> bl.load(BotLabeler())
    Views are fixed. Custom rules can be set locally (they are
    not inherited to other labelers). Rule config is accessible from
    all custom rules from ABCRule.config
    """

    def __init__(
        self,
        message_view: Optional["ABCBotMessageView"] = None,
        raw_event_view: Optional[RawBotEventView] = None,
        custom_rules: Optional[CustomRuleType] = None,
        auto_rules: Optional[List["ABCRule"]] = None,
        raw_event_auto_rules: Optional[List["ABCRule"]] = None,
    ) -> None:
        message_view = message_view or BotMessageView()
        raw_event_view = raw_event_view or RawBotEventView()
        super().__init__(
            message_view=message_view,
            raw_event_view=raw_event_view,
            custom_rules=custom_rules,
            auto_rules=auto_rules,
            raw_event_auto_rules=raw_event_auto_rules,
        )

    def message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        return super().message(*rules, blocking=blocking, **custom_rules)

    def chat_message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        return super().chat_message(*rules, blocking=blocking, **custom_rules)

    def private_message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        return super().private_message(*rules, blocking=blocking, **custom_rules)

    def raw_event(
        self,
        event: Union["EventName", List["EventName"]],
        dataclass: Union[Type[dict], Type["BaseGroupEvent"]] = dict,
        *rules: "ABCRule",
        blocking: bool = True,
        **custom_rules,
    ) -> "LabeledHandler":
        if any(not isinstance(rule, ABCRule) for rule in rules):
            msg = "All rules must be subclasses of ABCRule or rule shortcuts (https://vkbottle.rtfd.io/ru/latest/high-level/handling/rules/)"
            raise ValueError(msg)

        event_types = event if isinstance(event, list) else [event]

        def decorator(func):
            for e in event_types:
                if isinstance(e, str):
                    e = GroupEventType(e)
                handler_basement = BotHandlerBasement(
                    dataclass,
                    FromFuncHandler(
                        func,
                        *rules,
                        *self.raw_event_auto_rules,
                        *self.get_custom_rules(custom_rules),
                        blocking=blocking,
                    ),
                )
                event_handlers = self.raw_event_view.handlers.setdefault(e, [])
                event_handlers.append(handler_basement)
            return func

        return decorator



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, List, Mapping, Optional, Type, Union

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from ssl import SSLContext
    from types import SimpleNamespace

    from aiohttp import (
        BaseConnector,
        BasicAuth,
        ClientRequest,
        ClientResponse,
        ClientTimeout,
        ClientWebSocketResponse,
        Fingerprint,
        HttpVersion,
        TraceConfig,
    )
    from aiohttp.abc import AbstractCookieJar
    from aiohttp.typedefs import LooseCookies, LooseHeaders, StrOrURL

    from vkbottle.modules import JSONModule


class AiohttpSessionKwargs(TypedDict, total=False):
    """https://github.com/aio-libs/aiohttp/blob/33953f110e97eecc707e1402daa8d543f38a189b/aiohttp/client.py#L199"""

    base_url: Optional[StrOrURL]
    connector: Optional[BaseConnector]
    loop: Optional[AbstractEventLoop]
    cookies: Optional[LooseCookies]
    headers: Optional[LooseHeaders]
    skip_auto_headers: Optional[Iterable[str]]
    auth: Optional[BasicAuth]

    # Пришлось изменить из-за реализации AiohttpClient
    json_serialize: JSONModule

    request_class: Type[ClientRequest]
    response_class: Type[ClientResponse]
    ws_response_class: Type[ClientWebSocketResponse]
    version: HttpVersion
    cookie_jar: Optional[AbstractCookieJar]
    connector_owner: bool
    raise_for_status: bool
    read_timeout: Union[float, object]
    conn_timeout: Optional[float]
    timeout: Union[object, ClientTimeout]
    auto_decompress: bool
    trust_env: bool
    requote_redirect_url: bool
    trace_configs: Optional[List[TraceConfig]]
    read_bufsize: int


class AiohttpRequestKwargs(TypedDict, total=False):
    """https://github.com/aio-libs/aiohttp/blob/33953f110e97eecc707e1402daa8d543f38a189b/aiohttp/client.py#L369"""

    params: Optional[Mapping[str, str]]

    # Методы ABCHTTPClient уже принимают data
    # data: Any

    json: Any
    cookies: Optional[LooseCookies]
    headers: Optional[LooseHeaders]
    skip_auto_headers: Optional[Iterable[str]]
    auth: Optional[BasicAuth]
    allow_redirects: bool
    max_redirects: int
    compress: Optional[str]
    chunked: Optional[bool]
    expect100: bool
    raise_for_status: Optional[bool]
    read_until_eof: bool
    proxy: Optional[StrOrURL]
    proxy_auth: Optional[BasicAuth]
    timeout: Union[ClientTimeout, object]
    verify_ssl: Optional[bool]
    fingerprint: Optional[bytes]
    ssl_context: Optional[SSLContext]
    ssl: Optional[Union[SSLContext, bool, Fingerprint]]
    proxy_headers: Optional[LooseHeaders]
    trace_request_ctx: Optional[SimpleNamespace]
    read_bufsize: Optional[int]



import sys
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

if TYPE_CHECKING:
    from types import ModuleType

T_CodeException = TypeVar("T_CodeException", bound="CodeException")


def get_code_exception(cls: Type[T_CodeException], code: int) -> Type[T_CodeException]:
    return CodeExceptionFactory(cls, code).get()


class CodeException(Exception):
    code: ClassVar[int]

    def __init_subclass__(cls, code: Optional[int] = None, **kwargs: Dict[str, object]):
        super().__init_subclass__(**kwargs)
        if code is not None:
            cls.code = code
            for base in cls.mro()[1:]:
                if issubclass(base, CodeException):
                    sys.modules[base.__module__].__dict__[f"{base.__name__}_{code}"] = cls
                    break

    @overload
    def __class_getitem__(
        cls: Type[T_CodeException],
        code_or_codes: int,
    ) -> Type[T_CodeException]: ...

    @overload
    def __class_getitem__(
        cls: Type[T_CodeException],
        code_or_codes: Tuple[int, ...],
    ) -> Tuple[Type[T_CodeException], ...]: ...

    def __class_getitem__(
        cls: Type[T_CodeException],
        code_or_codes: Union[int, Tuple[int, ...]],
    ) -> Union[Type[T_CodeException], Tuple[Type[T_CodeException], ...]]:
        if isinstance(code_or_codes, int):
            return get_code_exception(cls, code_or_codes)
        return tuple(get_code_exception(cls, code) for code in code_or_codes)


class CodeExceptionFactory(Generic[T_CodeException]):
    def __init__(self, cls: Type[T_CodeException], code: int):
        self._class = cls
        self._code = code

    def get(self) -> Type[T_CodeException]:
        if not self._code_exists:
            self._create_code_exception()
        return self._code_exception

    @property
    def _code_exists(self) -> bool:
        return self._code_exception_name in self._base_class_namespace

    def _create_code_exception(self) -> None:
        code_exception: Type[CodeException] = type(
            self._code_exception_name, (self._class,), {"__module__": self._class.__module__}
        )
        code_exception.code = self._code
        self._base_class_namespace[self._code_exception_name] = code_exception

    @property
    def _code_exception(self) -> Type[T_CodeException]:
        return self._base_class_namespace[self._code_exception_name]

    @property
    def _code_exception_name(self) -> str:
        return f"{self._class.__name__}_{self._code}"

    @property
    def _base_class_namespace(self) -> Dict[str, Any]:
        return self._base_class_module.__dict__

    @property
    def _base_class_module(self) -> "ModuleType":
        return sys.modules[self._class.__module__]



from typing import TYPE_CHECKING, List, Optional

from vkbottle.exception_factory import CaptchaError, VKAPIError

if TYPE_CHECKING:
    from vkbottle.http import ABCHTTPClient

MOBILE_APP_ID = 2274003
MOBILE_APP_SECRET = "hHbZxrka2uZ6jB1inYsH"  # noqa: S105


class AuthError(VKAPIError[0]):  # type: ignore
    def __init__(
        self,
        *,
        error_description: str,
        error_type: str,
        error_msg: str,
        request_params: Optional[List[dict]] = None,
    ):
        request_params = request_params or []
        super().__init__(error_msg=error_msg, request_params=request_params)
        self.error_msg = error_msg
        self.error_type = error_type
        self.error_description = error_description


class UserAuth:
    AUTH_URL = "https://oauth.vk.com/token"

    def __init__(
        self,
        client_id: Optional[int] = None,
        client_secret: Optional[str] = None,
        http_client: Optional["ABCHTTPClient"] = None,
    ):
        from vkbottle.http import SingleAiohttpClient

        if client_id is not None and client_secret is not None:
            self.client_id = client_id
            self.client_secret = client_secret
        else:
            self.client_id = MOBILE_APP_ID
            self.client_secret = MOBILE_APP_SECRET

        self.http_client = http_client or SingleAiohttpClient()

    def _get_params(self, login: str, password: str) -> dict:
        return {
            "grant_type": "password",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": login,
            "password": password,
            "scope": 501202911,
        }

    async def get_token(self, login: str, password: str) -> str:
        params = self._get_params(login, password)

        response = await self.http_client.request_json(
            url=self.AUTH_URL,
            method="POST",
            data=params,
        )

        if "access_token" in response:
            return response["access_token"]
        response["error_msg"] = response.pop("error")
        if response["error_msg"] == "need_captcha":
            raise CaptchaError(**response, request_params=[])
        raise AuthError(**response, request_params=[])


__all__ = ("AuthError", "UserAuth")



import asyncio
import datetime
import typing

from vkbottle.dispatch.rules.abc import ABCRule
from vkbottle.dispatch.views.abc import ABCDispenseView
from vkbottle.tools.limited_dict import LimitedDict

from .middleware import WaiterMiddleware
from .short_state import Behaviour, Event, ShortState

Identificator = typing.Union[str, int]
Storage = typing.Dict[str, LimitedDict[Identificator, "ShortState[Event]"]]


class WaiterMachine:
    def __init__(self, *, max_storage_size: int = 1000) -> None:
        self.storage: Storage = {}
        self.max_storage_size = max_storage_size
        self.middleware = WaiterMiddleware(self)

    async def drop(
        self,
        dispensable_view: ABCDispenseView[dict, Event],
        identifier: Identificator,
        **context: typing.Any,
    ) -> None:
        view_name = dispensable_view.__class__.__name__
        if view_name not in self.storage:
            msg = f"No record of view {view_name} found"
            raise LookupError(msg)

        short_state = self.storage[view_name].pop(identifier, None)
        if not short_state:
            msg = f"Waiter with identificator {identifier} is not found for view {view_name}"
            raise LookupError(msg)

        if short_state.context is not None:
            context.update(short_state.context.context)

        short_state.cancel()
        await self.call_behaviour(
            dispensable_view,  # type: ignore
            short_state.on_drop_behaviour,
            short_state.event,
            **context,
        )

    async def wait(
        self,
        dispensable_view: ABCDispenseView[dict, Event],
        linked_event: Event,
        *rules: ABCRule[Event],
        default: Behaviour[Event] = None,
        on_drop: Behaviour[Event] = None,
        exit: Behaviour[Event] = None,  # noqa: A002
        expiration: typing.Union[datetime.timedelta, float, None] = None,
    ):
        if isinstance(expiration, (int, float)):
            expiration = datetime.timedelta(seconds=expiration)

        key = dispensable_view.get_state_key(linked_event)
        if not key:
            msg = "Unable to get state key."
            raise RuntimeError(msg)

        event = asyncio.Event()
        short_state = ShortState(
            key,
            ctx_api=linked_event.ctx_api,  # type: ignore
            event=event,
            rules=rules,
            expiration=expiration,
            default_behaviour=default,
            on_drop_behaviour=on_drop,
            exit_behaviour=exit,
        )

        view_name = dispensable_view.__class__.__name__
        if view_name not in self.storage:
            dispensable_view.middlewares.insert(0, self.middleware)  # type: ignore
            self.storage[view_name] = LimitedDict(maxlimit=self.max_storage_size)

        if (deleted_short_state := self.storage[view_name].set(key, short_state)) is not None:
            deleted_short_state.cancel()

        self.storage[view_name][key] = short_state
        await event.wait()
        self.storage[view_name].pop(key)

        if short_state.context is None:
            msg = "Context is not defined."
            raise RuntimeError(msg)

        return short_state.context

    async def call_behaviour(
        self,
        view: ABCDispenseView[dict, Event],
        behaviour: Behaviour[Event],
        event: Event,
        **context: typing.Any,
    ) -> bool:
        if behaviour is None:
            return False

        result = await behaviour.filter(event)
        if result is not False:
            result = result if isinstance(result, dict) else {}
            context.update(result)
            value = await behaviour.handle(event, **result)
            if return_handler := view.handler_return_manager.get_handler(value):
                await return_handler(
                    view.handler_return_manager,
                    value,
                    event,
                    context,
                )
            return True

        return False



from typing import (
    ClassVar,
    Dict,
    Optional,
    Tuple,
    TypeVar,
    overload,
)

T_CodeExceptionMeta = TypeVar("T_CodeExceptionMeta", bound="_CodeExceptionMeta")

T_CodeException = TypeVar("T_CodeException", bound="CodeException")

class _CodeExceptionMeta(type):
    code: int

    @overload
    def __getitem__(cls: T_CodeExceptionMeta, code_or_codes: int) -> T_CodeExceptionMeta: ...
    @overload
    def __getitem__(
        cls: T_CodeExceptionMeta, code_or_codes: Tuple[int, ...]
    ) -> Tuple[T_CodeExceptionMeta, ...]: ...

class CodeException(Exception, metaclass=_CodeExceptionMeta):
    code: ClassVar[int]

    def __init_subclass__(cls, code: Optional[int] = None, **kwargs: Dict[str, object]): ...



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, List, NoReturn, Optional, Type, TypeVar, Union

if TYPE_CHECKING:
    from vkbottle.dispatch.handlers.abc import ABCHandler
    from vkbottle.dispatch.views.abc import ABCView


T = TypeVar("T")


class MiddlewareError(Exception):
    pass


class ABCMiddleware(ABC):
    @abstractmethod
    def stop(self, error: Any) -> NoReturn: ...

    @abstractmethod
    def send(self, context_update: Optional[dict] = None) -> None: ...

    @abstractmethod
    async def pre(self) -> None: ...

    @abstractmethod
    async def post(self) -> None: ...

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"


class BaseMiddleware(Generic[T]):
    event: T
    view: Optional["ABCView"]
    handle_responses: List[Any]
    handlers: List["ABCHandler"]

    def __init__(self, event: T, view: Optional["ABCView"] = None):
        self.event = event
        self.view = view

        self.handle_responses = []
        self.handlers = []

        self._new_context: dict = {}
        self.error: Optional[Exception] = None

        self.pre = self.catch_all(self.pre)  # type: ignore
        self.post = self.catch_all(self.post)  # type: ignore

    def get_handle_response(self, handler) -> Optional[Any]:
        """Get handle response value for handler"""
        for handler_, response in zip(self.handlers, self.handle_responses):
            if handler_ == handler:
                return response

    @property
    def can_forward(self) -> bool:
        """Check if the event can be further processed"""

        return not self.error

    @property
    def context_update(self) -> dict:
        return self._new_context

    def catch_all(self, func):
        """Catch any exception and save error value"""

        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                self.error = e

        return wrapper

    def stop(self, error: Union[str, Exception, Type[Exception], None] = None) -> NoReturn:
        """Wrapper for exception raise"""

        if error is None or isinstance(error, str):
            raise MiddlewareError(error or "")
        raise error

    def send(self, context_update: Optional[dict] = None) -> None:
        """Validate new context update data if needed"""

        if context_update is not None:
            if not isinstance(context_update, dict):
                msg = "Context update value should be an instance of dict"
                raise ValueError(msg)
            self._new_context.update(context_update)

    async def pre(self) -> None: ...

    async def post(self) -> None: ...



from typing import List, Optional, Union

from vkbottle.modules import json
from vkbottle.tools.keyboard import Keyboard
from vkbottle.tools.keyboard.button import KeyboardButton


class TemplateElement:
    """Easy template element generator"""

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        photo_id: Optional[str] = None,
        buttons: Union[List[KeyboardButton], List[dict], str, None] = None,
        action: Optional[dict] = None,
    ):
        if not buttons:
            msg = "Buttons are required"
            raise ValueError(msg)
        if not (photo_id or title):
            msg = "photo_id or title are required"
            raise ValueError(msg)

        if isinstance(buttons, str):
            buttons = json.loads(buttons)["buttons"][0]  # type: ignore

        if isinstance(buttons, dict):
            buttons = buttons["buttons"][0]  # type: ignore

        if isinstance(buttons, Keyboard):
            buttons = [button.get_data() for button in buttons.buttons[0]]

        if all(isinstance(button, KeyboardButton) for button in buttons):  # type: ignore
            buttons = [button.get_data() for button in buttons]  # type: ignore
        self.raw: dict = {k: v for k, v in locals().items() if v is not None and k != "self"}



import ast
from inspect import getsource
from typing import Callable, Dict, Type


class ConverterError(Exception):
    pass


class Converter:
    """Translate Python into the VKScript with AST"""

    def __init__(self):
        self.definitions: Dict[Type[ast.AST], Callable] = {}

    def __call__(self, for_definition: Type[ast.AST]):
        def decorator(func):
            self.definitions[for_definition] = func
            return func

        return decorator

    def find_definition(self, d):
        if d.__class__ not in self.definitions:
            msg = f"Definition for {d.__class__} is undefined. Maybe vkscript doesn't support it"
            raise ConverterError(msg)
        return self.definitions[d.__class__](d)

    def scriptify(self, func: Callable, *args_values, **kwargs_values) -> str:
        """Translate function to VKScript"""
        source = getsource(func)
        code: ast.FunctionDef = ast.parse(source).body[0]  # type: ignore
        # Check if function has *args or **kwargs
        if code.args.vararg or code.args.kwarg:
            msg = "VKScript converter doesn't support *args and **kwargs"
            raise ConverterError(msg)
        # Get list of function arguments names
        args = [a.arg for a in code.args.args]
        # Get list of function arguments default values
        defaults = [self.find_definition(d) for d in code.args.defaults]
        # Check that first argument is api
        if not args or args[0] != "api":
            msg = "First argument must be api"
            raise ConverterError(msg)
        # Remove api from args
        args = args[1:]
        # Cycle through function arguments and check if they values are passed
        for arg, v in zip(args, args_values):
            if arg in kwargs_values:
                continue
            kwargs_values[arg] = v
        # Default values are used if values are not passed
        # eg func(a, b=2, c=3) -> func(1, 3), args = [a=1, b=3, c=3], defaults = [2, 3]
        for arg in args[::-1]:
            if not defaults and arg not in kwargs_values:
                msg = f"Argument {arg} is not provided"
                raise ConverterError(msg)
            if arg in kwargs_values:
                continue
            kwargs_values[arg] = defaults.pop()
        # Create assignments for every argument as variable
        values_assignments = [f"var {k}={v};" for k, v in kwargs_values.items()]
        return "".join(values_assignments) + "".join(
            self.find_definition(line) for line in code.body
        )



from typing import TYPE_CHECKING, List, Optional, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class PhotoUploader(BaseUploader):
    NAME = "picture.jpg"

    @property
    def attachment_name(self) -> str:
        return self._attachment_name or self.NAME


class PhotoToAlbumUploader(PhotoUploader):
    MAX_PHOTOS_PER_UPLOAD = 5

    async def upload(
        self,
        album_id: int,
        paths_like: Union[List[Union[str, "Bytes"]], str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> List[str]:
        photos = await self.raw_upload(
            album_id=album_id,
            paths_like=paths_like,
            group_id=group_id,
            **params,
        )
        return [
            self.generate_attachment_string(
                "photo",
                photo["owner_id"],
                photo["id"],
                photo.get("access_key"),
            )
            for photo in photos
        ]

    async def raw_upload(
        self,
        album_id: int,
        paths_like: Union[List[Union[str, "Bytes"]], str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> List[dict]:
        if not isinstance(paths_like, list):
            paths_like = [paths_like]
        if len(paths_like) > self.MAX_PHOTOS_PER_UPLOAD:
            msg = "You can upload up to 5 photos at once"
            raise ValueError(msg)
        server = await self.get_server(album_id=album_id, group_id=group_id, **params)
        files = {}

        for i, file_source in enumerate(paths_like):
            data = await self.read(file_source)
            files[f"file{i+1}"] = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], files)
        return (
            await self.api.request("photos.save", {"album_id": album_id, **uploader, **params})
        )["response"]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getUploadServer", kwargs))["response"]


class PhotoWallUploader(PhotoUploader):
    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> str:
        photo = await self.raw_upload(
            file_source=file_source,
            group_id=group_id,
            **params,
        )
        return self.generate_attachment_string(
            "photo",
            photo["owner_id"],
            photo["id"],
            photo.get("access_key"),
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: Optional[int] = None,
        **params,
    ) -> dict:
        server = await self.get_server(group_id=group_id, **params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        photos = (await self.api.request("photos.saveWallPhoto", {**uploader, **params}))[
            "response"
        ]

        return photos[0]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getWallUploadServer", kwargs))["response"]


class PhotoFaviconUploader(PhotoUploader):
    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        owner_id: Optional[int] = None,
        **params,
    ) -> str:
        owner_id = owner_id or await self.get_owner_id(**params)
        photo = await self.raw_upload(
            file_source=file_source,
            owner_id=owner_id,
            **params,
        )
        return self.generate_attachment_string(
            "wall",
            owner_id,
            photo["post_id"],
            photo.get("access_key"),
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        owner_id: Optional[int] = None,
        **params,
    ) -> dict:
        owner_id = owner_id or await self.get_owner_id(**params)
        server = await self.get_server(owner_id=owner_id)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        return (await self.api.request("photos.saveOwnerPhoto", {**uploader, **params}))[
            "response"
        ]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getOwnerPhotoUploadServer", kwargs))["response"]


class PhotoMessageUploader(PhotoUploader):
    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        peer_id: Optional[int] = None,
        **params,
    ) -> str:
        photo = await self.raw_upload(
            file_source=file_source,
            peer_id=peer_id,
            **params,
        )
        return self.generate_attachment_string(
            "photo",
            photo["owner_id"],
            photo["id"],
            photo.get("access_key"),
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        peer_id: Optional[int] = None,
        **params,
    ) -> dict:
        server = await self.get_server(peer_id=peer_id, **params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        return (
            await self.api.request(
                "photos.saveMessagesPhoto",
                {**uploader, **params},
            )
        )[
            "response"
        ][0]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getMessagesUploadServer", kwargs))["response"]


class PhotoChatFaviconUploader(PhotoUploader):
    async def upload(
        self,
        **params,
    ) -> str:
        msg = f"{self.__class__.__name__} does not support upload() method. Use raw_upload(...) instead."
        raise NotImplementedError(msg)

    async def raw_upload(
        self,
        chat_id: int,
        file_source: Union[str, "Bytes"],
        crop_x: Optional[int] = None,
        crop_y: Optional[int] = None,
        crop_width: Optional[int] = None,
        **params,
    ) -> dict:
        server = await self.get_server(
            chat_id=chat_id,
            crop_x=crop_x,
            crop_y=crop_y,
            crop_width=crop_width,
            **params,
        )
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        return await self.api.request(
            "photos.setChatPhoto", {"file": uploader["response"], **params}
        )

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getChatUploadServer", kwargs))["response"]


class PhotoMarketUploader(PhotoUploader):
    async def upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: int,
        **params,
    ) -> str:
        photo = await self.raw_upload(
            file_source=file_source,
            group_id=group_id,
            **params,
        )
        return self.generate_attachment_string(
            "photo",
            photo["owner_id"],
            photo["id"],
            photo.get("access_key"),
        )

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        group_id: int,
        **params,
    ) -> dict:
        server = await self.get_server(group_id=group_id, **params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"file": file})
        return await self.api.request(
            "photos.saveMarketPhoto",
            {**uploader, **params},
        )

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getMarketUploadServer", kwargs))["response"]


__all__ = (
    "PhotoChatFaviconUploader",
    "PhotoFaviconUploader",
    "PhotoMarketUploader",
    "PhotoMessageUploader",
    "PhotoToAlbumUploader",
    "PhotoUploader",
    "PhotoWallUploader",
)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from aiohttp import ClientSession

from vkbottle.modules import json as json_module
from vkbottle.tools.singleton import ABCSingleton

from .abc import ABCHTTPClient

if TYPE_CHECKING:
    from types import TracebackType

    from aiohttp import ClientResponse
    from typing_extensions import Unpack

    from vkbottle.modules import JSONModule

    from .aiohttp_types import AiohttpRequestKwargs, AiohttpSessionKwargs


class AiohttpClient(ABCHTTPClient):
    def __init__(
        self,
        session: Optional[ClientSession] = None,
        json_processing_module: Optional[JSONModule] = None,
        optimize: bool = False,
        **session_params: Unpack[AiohttpSessionKwargs],
    ) -> None:
        json_serialize = session_params.pop("json_serialize", None)
        self.json_processing_module = json_processing_module or json_serialize or json_module

        if optimize:
            session_params["skip_auto_headers"] = {"User-Agent"}
            session_params["raise_for_status"] = True

        self.session = session

        self._session_params = session_params

    async def request_raw(  # type: ignore[override]
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Unpack[AiohttpRequestKwargs],
    ) -> ClientResponse:
        if not self.session:
            self.session = ClientSession(  # type: ignore[misc]
                json_serialize=self.json_processing_module.dumps,
                **self._session_params,  # type: ignore[arg-type]
            )
        async with self.session.request(url=url, method=method, data=data, **kwargs) as response:
            await response.read()
            return response

    async def request_json(  # type: ignore[override]
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Unpack[AiohttpRequestKwargs],
    ) -> dict:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8", loads=self.json_processing_module.loads, content_type=None
        )

    async def request_text(  # type: ignore[override]
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Unpack[AiohttpRequestKwargs],
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text(encoding="utf-8")

    async def request_content(  # type: ignore[override]
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Unpack[AiohttpRequestKwargs],
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    def __del__(self) -> None:
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector._close()
            self.session._connector = None


class SingleAiohttpClient(AiohttpClient, ABCSingleton):
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        pass  # no need to close session in this case



from typing_extensions import deprecated  # type: ignore

from vkbottle.framework.labeler import BotLabeler

from .dispatch.rules import base as rules
from .dispatch.views import bot as views
from .framework.bot import Bot, BotBlueprint, run_multibot
from .tools.mini_types.bot import MessageEventMin, MessageMin

Message = MessageMin
MessageEvent = MessageEventMin


@deprecated(
    "Blueprints was deprecated and will be removed in future releases, "
    "read about new code separation method in documentation: \n"
    "https://vkbottle.rtfd.io/ru/latest/tutorial/code-separation/",
    stacklevel=0,
)
class Blueprint(BotBlueprint): ...


__all__ = (
    "Blueprint",
    "Bot",
    "BotLabeler",
    "Message",
    "MessageEvent",
    "rules",
    "run_multibot",
    "views",
)



from abc import ABC, abstractmethod
from typing import Any, Callable, Tuple, Type, Union


class ABCValidator(ABC):
    @abstractmethod
    async def check(self, value: Any) -> bool:
        pass


class IsInstanceValidator(ABCValidator):
    def __init__(self, t: Union[Type[Any], Tuple[Type[Any], ...]]):
        self.t = t

    async def check(self, value: Any) -> bool:
        return isinstance(value, self.t)


class EqualsValidator(ABCValidator):
    def __init__(self, primary_value: Any):
        self.primary_value = primary_value

    async def check(self, value: Any) -> bool:
        return value == self.primary_value


class CallableValidator(ABCValidator):
    def __init__(self, call: Callable[[Any], bool]):
        self.call = call

    async def check(self, value: Any) -> bool:
        return self.call(value)


__all__ = ("ABCValidator", "IsInstanceValidator", "EqualsValidator", "CallableValidator")



from typing import TYPE_CHECKING, List, Optional

from vkbottle.modules import pydantic
from vkbottle.tools.mini_types.base import BaseMessageMin

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI


from .foreign_message import ForeignMessageMin  # noqa: TCH001


class MessageMin(BaseMessageMin):
    user_id: Optional[int] = None
    reply_message: Optional["ForeignMessageMin"] = None
    fwd_messages: List["ForeignMessageMin"] = pydantic.Field(default_factory=list)

    @property
    def is_mentioned(self) -> bool:
        return self.mention.id == self.user_id if self.mention else False

    @pydantic.root_validator(pre=True)
    def __foreign_messages(cls, values):
        foreign_messages = []
        if values.get("fwd_messages"):
            foreign_messages.extend(values["fwd_messages"])
        if values.get("reply_message"):
            foreign_messages.append(values["reply_message"])
        for foreign_message in foreign_messages:
            foreign_message["unprepared_ctx_api"] = values["unprepared_ctx_api"]
            foreign_message["replace_mention"] = values["replace_mention"]
            foreign_message["user_id"] = values.get("user_id")
        return values


MessageMin.update_forward_refs()


async def message_min(
    message_id: int,
    ctx_api: "ABCAPI",
    replace_mention: bool = True,
) -> "MessageMin":
    response = await ctx_api.messages.get_by_id(message_ids=[message_id])
    if not response.items:
        msg = f"Message with id {message_id} not found, perhaps it was deleted"
        raise ValueError(msg)
    message_object = response.items[0].dict()
    return MessageMin(
        **message_object,
        unprepared_ctx_api=ctx_api,
        replace_mention=replace_mention,
    )



import dataclasses
from typing import TYPE_CHECKING, Dict, List, Type, Union

from vkbottle_types.events import GroupEventType

from vkbottle.dispatch.return_manager.bot import BotMessageReturnHandler
from vkbottle.dispatch.views.abc.raw import ABCRawEventView, BaseHandlerBasement

if TYPE_CHECKING:
    from vkbottle_types.events.bot_events import BaseGroupEvent

    from vkbottle.dispatch.handlers import ABCHandler


@dataclasses.dataclass(frozen=True)
class BotHandlerBasement(BaseHandlerBasement):
    dataclass: Union[Type[dict], Type["BaseGroupEvent"]]
    handler: "ABCHandler"


class RawBotEventView(ABCRawEventView[dict, BotHandlerBasement]):
    handlers: Dict[GroupEventType, List[BotHandlerBasement]]

    def __init__(self) -> None:
        super().__init__()
        self.handlers = {}
        self.handler_return_manager = BotMessageReturnHandler()

    def get_handler_basements(self, event: dict) -> List[BotHandlerBasement]:
        return self.handlers[GroupEventType(self.get_event_type(event))]

    def get_event_model(
        self,
        handler_basement: BotHandlerBasement,
        event: dict,
    ) -> Union[dict, "BaseGroupEvent"]:
        return handler_basement.dataclass(**event)

    @staticmethod
    def get_event_type(event: dict) -> str:
        return event["type"]

    async def process_event(self, event: dict) -> bool:
        return GroupEventType(self.get_event_type(event)) in self.handlers



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, AsyncIterator, Optional

from typing_extensions import Self

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.exception_factory import ABCErrorHandler


class ABCPolling(ABC):
    """Abstract Polling class
    Documentation: https://vkbottle.rtfd.io/ru/latest/low-level/polling
    """

    @abstractmethod
    async def get_server(self) -> Any:
        pass

    @abstractmethod
    async def get_event(self, server: Any) -> dict:
        pass

    @abstractmethod
    async def listen(self) -> AsyncIterator[dict]:
        yield {}

    @property
    @abstractmethod
    def api(self) -> "ABCAPI":
        pass

    @api.setter  # noqa: B027
    def api(self, new_api: "ABCAPI") -> None:
        pass

    @abstractmethod
    def construct(
        self,
        api: "ABCAPI",
        error_handler: Optional["ABCErrorHandler"] = None,
    ) -> Self:
        pass



from typing import TYPE_CHECKING, Dict

from typing_extensions import Self

from vkbottle.modules import logger

from .abc import ABCRouter

if TYPE_CHECKING:
    from vkbottle.api.abc import ABCAPI
    from vkbottle.dispatch.dispenser import ABCStateDispenser
    from vkbottle.dispatch.views import ABCView
    from vkbottle.exception_factory import ABCErrorHandler


class Router(ABCRouter):
    async def route(self, event: dict, ctx_api: "ABCAPI") -> None:
        logger.debug("Routing update {!r}", event)

        for view in self.views.values():
            try:
                if not await view.process_event(event):
                    continue
                await view.handle_event(event, ctx_api, self.state_dispenser)
            except Exception as e:
                await self.error_handler.handle(e)

    def construct(
        self,
        views: Dict[str, "ABCView"],
        state_dispenser: "ABCStateDispenser",
        error_handler: "ABCErrorHandler",
    ) -> Self:
        self.views = views
        self.state_dispenser = state_dispenser
        self.error_handler = error_handler
        return self



from typing import TYPE_CHECKING, Literal, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class SpeechUploader(BaseUploader):
    NAME = "voice.ogg"

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("asr.getUploadUrl", kwargs))["response"]

    async def raw_upload(
        self,
        file_source: Union[str, "Bytes"],
        model: Literal["neutral", "spontaneous"] = "neutral",
        **params,
    ) -> dict:
        server = await self.get_server(**params)
        data = await self.read(file_source)

        file = self.get_bytes_io(data, name=self.attachment_name)

        audio = await self.upload_files(server["upload_url"], {"file": file})

        return (
            await self.api.request(
                "asr.process",
                {"model": model, "audio": audio, **params},
            )
        )["response"]

    async def upload(
        self,
        **params,
    ) -> str:
        msg = f"{self.__class__.__name__} does not support upload() method. Use raw_upload(...) instead."
        raise NotImplementedError(msg)

    @property
    def attachment_name(self) -> str:
        return self._attachment_name or self.NAME



from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Awaitable, Callable, Dict, NamedTuple, Optional, Tuple, Union

HANDLER_PROPERTY_TYPES = Union[type, Tuple[type, ...]]


class HandlerProperty(NamedTuple):
    types: HANDLER_PROPERTY_TYPES
    handler: Callable


class ABCReturnManager(ABC):
    @abstractmethod
    def get_handler(self, value: Any) -> Any: ...

    @abstractproperty
    def handlers(self) -> Any: ...

    @classmethod
    @abstractmethod
    def instance_of(cls, types: Any) -> Any: ...

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"


class BaseReturnManager(ABCReturnManager):
    def get_handler(self, value: Any) -> Optional[Callable]:
        for types, handler in self.handlers.items():
            if isinstance(value, types):
                return handler

    @property
    def handlers(self) -> Dict[HANDLER_PROPERTY_TYPES, Callable[[Any], Awaitable]]:
        return {
            v.types: v.handler
            for k, v in vars(self.__class__).items()
            if isinstance(v, HandlerProperty)
        }

    @classmethod
    def instance_of(
        cls, types: Union[type, Tuple[type, ...]]
    ) -> Callable[[Callable], HandlerProperty]:
        def decorator(func: Callable):
            return HandlerProperty(types, func)

        return decorator



from typing import TYPE_CHECKING, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class AudioUploader(BaseUploader):
    NAME = "audio.mp3"

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("audio.getUploadServer", {}))["response"]

    async def upload(
        self,
        artist: str,
        title: str,
        file_source: Union[str, "Bytes"],
        **params,
    ) -> str:
        audio = await self.raw_upload(artist, title, file_source, **params)
        return self.generate_attachment_string(
            "audio",
            await self.get_owner_id(**params, **audio),
            audio["id"],
            audio.get("access_key"),
        )

    async def raw_upload(
        self,
        artist: str,
        title: str,
        file_source: Union[str, "Bytes"],
        **params,
    ) -> dict:
        server = await self.get_server()
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"file": file})

        return (
            await self.api.request(
                "audio.save",
                {"artist": artist, "title": title, **uploader, **params},
            )
        )["response"]

    @property
    def attachment_name(self) -> str:
        return self._attachment_name or self.NAME


__all__ = ("AudioUploader",)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


class ABCRequestRescheduler(ABC):
    @abstractmethod
    async def reschedule(
        self,
        ctx_api: Union["ABCAPI", "API"],
        method: str,
        data: dict,
        recent_response: Any,
    ) -> dict:
        pass



from abc import ABC, abstractmethod
from io import StringIO
from typing import TYPE_CHECKING, Any, Callable, List, Literal, Optional, Union, overload

from vkbottle_types.objects import (
    AudioAudio,
    DocsDoc,
    MessagesAudioMessage,
    MessagesForward,
    MessagesMessage,
    PhotosPhoto,
    UsersUserFull,
    VideoVideoFull,
    WallWallComment,
    WallWallpostFull,
)

from vkbottle.modules import json, logger, pydantic

if TYPE_CHECKING:
    from vkbottle_types.responses.messages import MessagesSendUserIdsResponseItem

    from vkbottle.api import ABCAPI, API


from vkbottle.dispatch.dispenser.base import StatePeer  # noqa: TCH001

from .foreign_message import BaseForeignMessageMin  # noqa: TCH001
from .mention import Mention, replace_mention_validator


class BaseMessageMin(MessagesMessage, ABC):
    unprepared_ctx_api: Optional[Any] = None
    state_peer: Optional["StatePeer"] = None
    reply_message: Optional["BaseForeignMessageMin"] = None
    fwd_messages: List["BaseForeignMessageMin"] = pydantic.Field(default_factory=list)
    replace_mention: Optional[bool] = None
    _mention: Optional[Mention] = None

    __replace_mention = pydantic.root_validator(
        replace_mention_validator, allow_reuse=True, pre=False
    )  # type: ignore

    class Config:
        frozen = False

    @property
    def ctx_api(self) -> Union["ABCAPI", "API"]:
        return self.unprepared_ctx_api  # type: ignore

    @property
    def mention(self) -> Optional[Mention]:
        """Returns `Mention` object if message contains mention,
        eg if message is `@username text` returns `Mention(id=123, text="text")`,
        also mention is automatically removes from message text"""
        if not self.replace_mention:
            logger.warning(
                "labeler.message_view.replace_mention is set to False, the mention will not be processed"
            )
            return None
        return self._mention

    @property
    @abstractmethod
    def is_mentioned(self) -> bool:
        """Returns True if current bot is mentioned in message"""
        ...

    @overload
    async def get_user(self, raw_mode: Literal[False] = ..., **kwargs) -> UsersUserFull: ...

    @overload
    async def get_user(self, raw_mode: Literal[True] = ..., **kwargs) -> dict: ...

    async def get_user(self, raw_mode: bool = False, **kwargs) -> Union[UsersUserFull, dict]:
        raw_user = (await self.ctx_api.request("users.get", {"user_ids": self.from_id, **kwargs}))[
            "response"
        ][0]
        return raw_user if raw_mode else UsersUserFull(**raw_user)

    @property
    def chat_id(self) -> int:
        return self.peer_id - 2_000_000_000

    @property
    def message_id(self) -> int:
        return self.conversation_message_id or self.id

    def get_attachment_strings(self) -> Optional[List[str]]:
        if self.attachments is None:
            return None
        attachments = []
        for attachment in self.attachments:
            attachment_type = attachment.type.value
            attachment_object = getattr(attachment, attachment_type)
            if not hasattr(attachment_object, "id") or not hasattr(attachment_object, "owner_id"):
                logger.debug("Got unsupported attachment type: {}", attachment_type)
                continue
            attachment_string = (
                f"{attachment_type}{attachment_object.owner_id}_{attachment_object.id}"
            )
            if hasattr(attachment_object, "access_key"):
                attachment_string += f"_{attachment_object.access_key}"
            attachments.append(attachment_string)
        return attachments

    def get_wall_attachment(self) -> Optional[List["WallWallpostFull"]]:
        if self.attachments is None:
            return None
        result = [attachment.wall for attachment in self.attachments if attachment.wall]
        return result or None

    def get_wall_reply_attachment(self) -> Optional[List["WallWallComment"]]:
        if self.attachments is None:
            return None
        result = [
            attachment.wall_reply for attachment in self.attachments if attachment.wall_reply
        ]
        return result or None

    def get_photo_attachments(self) -> Optional[List["PhotosPhoto"]]:
        if self.attachments is None:
            return None
        return [attachment.photo for attachment in self.attachments if attachment.photo]

    def get_video_attachments(self) -> Optional[List["VideoVideoFull"]]:
        if self.attachments is None:
            return None
        return [attachment.video for attachment in self.attachments if attachment.video]

    def get_doc_attachments(self) -> Optional[List["DocsDoc"]]:
        if self.attachments is None:
            return None
        return [attachment.doc for attachment in self.attachments if attachment.doc]

    def get_audio_attachments(self) -> Optional[List["AudioAudio"]]:
        if self.attachments is None:
            return None
        return [attachment.audio for attachment in self.attachments if attachment.audio]

    def get_audio_message_attachments(self) -> Optional[List["MessagesAudioMessage"]]:
        if self.attachments is None:
            return None
        return [
            attachment.audio_message for attachment in self.attachments if attachment.audio_message
        ]

    def get_message_id(self) -> Optional[int]:
        return self.id or self.conversation_message_id

    def get_payload_json(
        self,
        throw_error: bool = False,
        unpack_failure: Callable[[str], Union[dict, str]] = lambda payload: payload,
    ) -> Optional[Union[dict, str]]:
        if self.payload is None:
            return None
        try:
            return json.loads(self.payload)
        except (ValueError, TypeError) as e:
            if throw_error:
                raise e from e
        return unpack_failure(self.payload)

    async def answer(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        random_id: Optional[int] = 0,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        reply_to: Optional[int] = None,
        forward_messages: Optional[List[int]] = None,
        forward: Optional[str] = None,
        sticker_id: Optional[int] = None,
        keyboard: Optional[str] = None,
        template: Optional[str] = None,
        payload: Optional[str] = None,
        content_source: Optional[str] = None,
        dont_parse_links: Optional[bool] = None,
        disable_mentions: Optional[bool] = None,
        intent: Optional[str] = None,
        subscribe_id: Optional[int] = None,
        **kwargs,
    ) -> "MessagesSendUserIdsResponseItem":
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        deprecated_params = ("peer_id", "user_id", "domain", "chat_id", "user_ids")
        deprecated = [k for k in data if k in deprecated_params]
        if deprecated:
            logger.warning(
                "Params like peer_id or user_id is deprecated in Message.answer()."
                "Use API.messages.send() instead"
            )
            for k in deprecated:
                data.pop(k)
        if message is None:
            message = ""
        elif not isinstance(message, str):
            message = str(message)
        stream = StringIO(message)
        while True:
            msg = stream.read(4096)
            if msg:
                data["message"] = msg
            response = (await self.ctx_api.messages.send(peer_ids=[self.peer_id], **data))[0]  # type: ignore
            if stream.tell() == len(message or ""):
                break
        return response

    async def reply(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        **kwargs,
    ) -> "MessagesSendUserIdsResponseItem":
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        data["forward"] = MessagesForward(
            conversation_message_ids=[self.conversation_message_id],  # type: ignore
            peer_id=self.peer_id,
            is_reply=True,
        ).json()

        return await self.answer(**data)

    async def forward(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        **kwargs,
    ) -> "MessagesSendUserIdsResponseItem":
        locals().update(kwargs)

        data = {
            k: v
            for k, v in locals().items()
            if k not in ("self", "kwargs", "forward_message_ids") and v is not None
        }
        data["forward"] = MessagesForward(
            conversation_message_ids=[self.conversation_message_id],
            peer_id=self.peer_id,  # type: ignore
        ).json()

        return await self.answer(**data)


BaseMessageMin.update_forward_refs()



from typing_extensions import deprecated  # type: ignore

from vkbottle.framework.labeler import UserLabeler

from .dispatch.rules import base as rules
from .dispatch.views import user as views
from .framework.user import User, UserBlueprint, run_multibot
from .tools.mini_types.bot import MessageMin

Message = MessageMin


@deprecated(
    "Blueprints was deprecated and will be removed in future releases, "
    "read about new code separation method in documentation: \n"
    "https://vkbottle.rtfd.io/ru/latest/tutorial/code-separation/",
    stacklevel=0,
)
class Blueprint(UserBlueprint): ...


__all__ = (
    "Blueprint",
    "Message",
    "User",
    "UserLabeler",
    "rules",
    "run_multibot",
    "views",
)



from .abc import ABCStateDispenser
from .base import BaseStateGroup, StatePeer, get_state_repr
from .builtin import BuiltinStateDispenser

__all__ = (
    "ABCStateDispenser",
    "BaseStateGroup",
    "BuiltinStateDispenser",
    "StatePeer",
    "get_state_repr",
)



from typing import Optional, Union

Payload = Union[str, dict]


class ABCAction:
    type: str

    def get_data(self) -> dict:
        data = {k: v for k, v in vars(self).items() if v is not None}
        data["type"] = self.type
        return data


class Text(ABCAction):
    type = "text"

    def __init__(self, label: str, payload: Optional[Payload] = None):
        self.label = label
        self.payload = payload


class OpenLink(ABCAction):
    type = "open_link"

    def __init__(self, link: str, label: str, payload: Optional[Payload] = None):
        self.link = link
        self.label = label
        self.payload = payload


class Location(ABCAction):
    type = "location"

    def __init__(self, payload: Optional[Payload] = None):
        self.payload = payload


class VKPay(ABCAction):
    type = "vkpay"

    def __init__(
        self,
        payload: Optional[Payload] = None,
        hash: Optional[str] = None,  # noqa: A002
    ):
        self.payload = payload
        self.hash = hash


class VKApps(ABCAction):
    type = "open_app"

    def __init__(
        self,
        app_id: int,
        owner_id: int,
        payload: Optional[Payload] = None,
        label: Optional[str] = None,
        hash: Optional[str] = None,  # noqa: A002
    ):
        self.app_id = app_id
        self.owner_id = owner_id
        self.payload = payload
        self.label = label
        self.hash = hash


class Callback(ABCAction):
    type = "callback"

    def __init__(self, label: str, payload: Payload):
        self.label = label
        self.payload = payload


__all__ = (
    "ABCAction",
    "Callback",
    "Location",
    "OpenLink",
    "Text",
    "VKApps",
    "VKPay",
)



from abc import ABCMeta


class SingletonMeta(type):
    __instance__ = None

    def __call__(cls, *args, **kwargs):
        if cls.__instance__ is None:
            cls.__instance__ = super().__call__(*args, **kwargs)
        return cls.__instance__


class ABCSingletonMeta(ABCMeta, SingletonMeta):
    pass


class ABCSingleton(metaclass=ABCSingletonMeta):
    pass


__all__ = ("ABCSingleton", "ABCSingletonMeta", "SingletonMeta")



from abc import ABC, abstractmethod
from typing import List, Union


class ABCTokenGenerator(ABC):
    async def __aenter__(self):
        return await self.get_token()

    @abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    @abstractmethod
    async def get_token(self) -> str:
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__}>"


Token = Union[str, List[str], ABCTokenGenerator]



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Coroutine, Dict, Optional, Type, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")

T = TypeVar("T")
T_AsyncFunc = TypeVar("T_AsyncFunc", bound=Callable[..., Awaitable[object]])


class ABCErrorHandler(ABC):
    error_handlers: Dict[Type[Exception], Callable[..., Awaitable[Any]]]
    undefined_error_handler: Optional[Callable[..., Awaitable[Any]]]

    @abstractmethod
    def register_error_handler(
        self, *error_types: Type[Exception]
    ) -> Callable[[T_AsyncFunc], T_AsyncFunc]:
        pass

    @abstractmethod
    def register_undefined_error_handler(self, handler: T_AsyncFunc) -> T_AsyncFunc:
        pass

    @abstractmethod
    async def handle(self, error: Exception) -> Any:
        pass

    @abstractmethod
    def catch(self, func: Callable[P, Awaitable[T]]) -> Callable[P, Coroutine[Any, Any, T]]:
        pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from vkbottle.modules import logger

if TYPE_CHECKING:
    from .base import BaseStateGroup, StatePeer


class ABCStateDispenser(ABC):
    @abstractmethod
    async def get(self, peer_id: int) -> Optional["StatePeer"]:
        pass

    @abstractmethod
    async def set(self, peer_id: int, state: "BaseStateGroup", **payload):
        pass

    @abstractmethod
    async def delete(self, peer_id: int):
        pass

    async def cast(self, peer_id: Optional[int]) -> Optional["StatePeer"]:
        if peer_id is None:
            return None

        logger.debug("Casting state for peer_id {}", peer_id)
        return await self.get(peer_id)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, List, TypeVar

from vkbottle.modules import logger

from .dispense import ABCDispenseView

if TYPE_CHECKING:
    from vkbottle.api.abc import ABCAPI
    from vkbottle.dispatch.dispenser.abc import ABCStateDispenser
    from vkbottle.dispatch.handlers import ABCHandler
    from vkbottle.tools.mini_types.base import BaseMessageMin

T_contra = TypeVar("T_contra", list, dict, contravariant=True)
F_contra = TypeVar("F_contra", contravariant=True)

DEFAULT_STATE_KEY = "peer_id"


class ABCMessageView(ABCDispenseView[T_contra, F_contra], ABC, Generic[T_contra, F_contra]):
    handlers: List["ABCHandler[F_contra]"]
    state_source_key: str
    default_text_approximators: List[Callable[["BaseMessageMin"], str]]
    replace_mention = False

    def __init__(self):
        super().__init__()
        self.state_source_key = DEFAULT_STATE_KEY
        self.default_text_approximators = []

    @staticmethod
    @abstractmethod
    def get_event_type(event: T_contra) -> Any:
        pass

    @staticmethod
    @abstractmethod
    async def get_message(event: T_contra, ctx_api, replace_mention: bool) -> Any:
        pass

    async def handle_event(
        self, event: T_contra, ctx_api: "ABCAPI", state_dispenser: "ABCStateDispenser"
    ) -> None:
        # For user event mapping, consider checking out
        # https://dev.vk.com/api/user-long-poll/getting-started
        logger.debug("Handling event ({}) with message view", self.get_event_type(event))
        context_variables: dict = {}
        message = await self.get_message(event, ctx_api, self.replace_mention)
        message.state_peer = await state_dispenser.cast(self.get_state_key(message))

        for text_ax in self.default_text_approximators:
            message.text = text_ax(message)

        mw_instances = await self.pre_middleware(message, context_variables)
        if mw_instances is None:
            logger.debug("Handling stopped, pre_middleware returned error")
            return

        handle_responses = []
        handlers = []

        for handler in self.handlers:
            result = await handler.filter(message)
            logger.debug("Handler {} returned {}", handler, result)

            if result is False:
                continue

            elif isinstance(result, dict):
                context_variables.update(result)

            handler_response = await handler.handle(message, **context_variables)
            handle_responses.append(handler_response)
            handlers.append(handler)

            return_handler = self.handler_return_manager.get_handler(handler_response)
            if return_handler is not None:
                await return_handler(
                    self.handler_return_manager, handler_response, message, context_variables
                )

            if handler.blocking:
                break

        await self.post_middleware(mw_instances, handle_responses, handlers)



from typing import TYPE_CHECKING, NoReturn, Optional, Tuple

from vkbottle.api import API
from vkbottle.callback import BotCallback
from vkbottle.dispatch import BuiltinStateDispenser, Router
from vkbottle.exception_factory import ErrorHandler
from vkbottle.framework.abc import ABCFramework
from vkbottle.framework.labeler import BotLabeler
from vkbottle.modules import logger
from vkbottle.polling import BotPolling
from vkbottle.tools import LoopWrapper

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, Token
    from vkbottle.callback import ABCCallback
    from vkbottle.dispatch import ABCRouter, ABCStateDispenser
    from vkbottle.exception_factory import ABCErrorHandler
    from vkbottle.framework.labeler import ABCLabeler
    from vkbottle.polling import ABCPolling


class Bot(ABCFramework):
    def __init__(
        self,
        token: Optional["Token"] = None,
        api: Optional["ABCAPI"] = None,
        polling: Optional["ABCPolling"] = None,
        callback: Optional["ABCCallback"] = None,
        loop_wrapper: Optional[LoopWrapper] = None,
        router: Optional["ABCRouter"] = None,
        labeler: Optional["ABCLabeler"] = None,
        state_dispenser: Optional["ABCStateDispenser"] = None,
        error_handler: Optional["ABCErrorHandler"] = None,
        task_each_event=None,
    ) -> None:
        self.api: API = api or API(token)  # type: ignore
        self.error_handler = error_handler or ErrorHandler()
        self.loop_wrapper = loop_wrapper or LoopWrapper()
        self.labeler = labeler or BotLabeler()
        self.state_dispenser = state_dispenser or BuiltinStateDispenser()
        self._polling = polling or BotPolling(self.api)
        self._callback = callback or BotCallback()
        self._router = router or Router()
        if task_each_event:
            logger.warning("task_each_event is deprecated and will be removed in future versions")

    @property
    def callback(self) -> "ABCCallback":
        if self._callback is None:
            msg = "To work with this methods, you need to create a BotCallback class and pass it as a parameter to the Bot class"
            raise ValueError(msg)
        return self._callback.construct(self.api, self.error_handler)

    @property
    def polling(self) -> "ABCPolling":
        return self._polling.construct(self.api, self.error_handler)

    @property
    def router(self) -> "ABCRouter":
        return self._router.construct(
            views=self.labeler.views(),
            state_dispenser=self.state_dispenser,
            error_handler=self.error_handler,
        )

    @router.setter
    def router(self, new_router: "ABCRouter") -> None:
        self._router = new_router

    @property
    def on(self) -> "ABCLabeler":
        return self.labeler

    async def run_polling(self, custom_polling: Optional["ABCPolling"] = None) -> NoReturn:  # type: ignore
        polling = custom_polling or self.polling
        logger.info("Starting polling for {!r}", polling.api)

        async for event in polling.listen():
            logger.debug("New event was received: {!r}", event)
            for update in event["updates"]:
                self.loop_wrapper.add_task(self.router.route(update, polling.api))

    def run_forever(self) -> NoReturn:  # type: ignore
        logger.info("Loop will be run forever")
        self.loop_wrapper.add_task(self.run_polling())
        self.loop_wrapper.run()

    async def setup_webhook(self) -> Tuple[str, str]:
        """:return: confirmation_code, secret_key"""

        await self.callback.setup_group_id()

        confirmation_code: str = await self.callback.get_callback_confirmation_code()
        secret_key: str = self.callback.get_secret_key()

        server_id = await self.callback.find_server_id()

        if server_id is not None:
            await self.callback.set_callback_settings(server_id, {"message_new": True})
            await self.callback.edit_callback_server(server_id)
        else:
            server_id = await self.callback.add_callback_server()
            await self.callback.set_callback_settings(server_id, {"message_new": True})

        return confirmation_code, secret_key

    async def process_event(self, event: dict) -> None:
        await self.router.route(event, self.api)



from abc import ABC, abstractmethod


class ABCRequestValidator(ABC):
    """Abstract Request Validator class
    Documentation: https://vkbottle.rtfd.io/ru/latest/low-level/api/request_validator
    """

    @abstractmethod
    async def validate(self, request: dict) -> dict:
        pass



import ast
import random
import string
from typing import TYPE_CHECKING, Any, Callable, Iterable

from typing_extensions import Concatenate, ParamSpec

from .base_converter import Converter, ConverterError

if TYPE_CHECKING:
    from vkbottle import ABCAPI

CALL_REPLACEMENTS = {
    "append": "push",
    "pop": "pop",
    "sort": "sort",
    "index": "indexOf",
}
CALL_STRING = ["join", "strip", "split"]
CALL_TYPES = {
    "str": "toString",
}

converter = Converter()
find = converter.find_definition


def dispatch_keywords(keywords: Iterable, assigner: str = ":", sep: str = ","):
    return sep.join(f"{param.arg}{assigner}{find(param.value)}" for param in keywords)


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    return components[0].lower() + "".join(x.title() for x in components[1:])


@converter(ast.Assign)
def assign(d: ast.Assign):
    left = d.targets
    left_ = [find(target) for target in left if target.__class__ == ast.Name]
    right = find(d.value)
    return "var " + ",".join(f"{target}={right}" for target in left_) + ";"


@converter(ast.Add)
@converter(ast.UAdd)
def add_operator(_: ast.Add):
    return "+"


@converter(ast.Sub)
@converter(ast.USub)
def sub_operator(_: ast.Sub):
    return "-"


@converter(ast.Mult)
def mult_operator(_: ast.Mult):
    return "*"


@converter(ast.Div)
def div_operator(_: ast.Div):
    return "/"


@converter(ast.Pow)
def pow_operator(_: ast.Pow):
    return "**"


@converter(ast.RShift)
def rshift_operator(_: ast.RShift):
    return ">>"


@converter(ast.LShift)
def lshift_operator(_: ast.LShift):
    return "<<"


@converter(ast.BitAnd)
def bitand_operator(_: ast.BitAnd):
    return "&"


@converter(ast.BitOr)
def bitor_operator(_: ast.BitOr):
    return "|"


@converter(ast.Mod)
def mod_operator(_: ast.Mod):
    return "%"


@converter(ast.Gt)
def gt_operator(_: ast.Gt):
    return ">"


@converter(ast.Lt)
def lt_operator(_: ast.Lt):
    return "<"


@converter(ast.GtE)
def gte_operator(_: ast.GtE):
    return ">="


@converter(ast.LtE)
def lte_operator(_: ast.LtE):
    return "<="


@converter(ast.Eq)
def eq_operator(_: ast.Eq):
    return "=="


@converter(ast.Is)
def is_operator(_: ast.Is):
    return "==="


@converter(ast.NotEq)
def noteq_operator(_: ast.Gt):
    return "!="


@converter(ast.IsNot)
def isnot_operator(_: ast.IsNot):
    return "!=="


@converter(ast.And)
def and_operator(_: ast.And):
    return "&&"


@converter(ast.Or)
def or_operator(_: ast.Or):
    return "||"


@converter(ast.AugAssign)
def aug_assign(d: ast.AugAssign):
    operator = find(d.op)
    value = find(d.value)
    target = find(d.target)
    return f"{target} = {target} {operator} {value};"


@converter(ast.Constant)
def constant(d: ast.Constant):
    # Since python 3.8 ast.Constant is used
    # instead of ast.Num, ast.Str, ast.Bytes and ast.NameConstant
    if isinstance(d.value, bool):
        return "true" if d.value else "false"
    elif d.value is None:
        return "null"
    elif isinstance(d.value, str):
        return repr(d.value)
    elif isinstance(d.value, bytes):
        return repr(d.value.decode())
    elif isinstance(d.value, int):
        return str(d.value)
    return d.value


@converter(ast.Name)
def name(d: ast.Name):
    return d.id


@converter(ast.While)
def while_cycle(d: ast.While):
    if d.orelse:
        msg = "You can't use while or/else in vkscript"
        raise ConverterError(msg)
    body = "".join(find(line) for line in d.body)
    return f"while({find(d.test)}" + "){" + body + "};"


@converter(ast.For)
def for_cycle(d: ast.For):
    random_iter_name = f"__iter_{random_string(5)}__"
    body = "".join(find(line) for line in d.body)
    return (
        f"var {random_iter_name} = {find(d.iter)};"
        f"while({random_iter_name}.length > 0){{var {find(d.target)}={random_iter_name}.pop();{body}}};"
    )


@converter(ast.If)
def if_statement(d: ast.If):
    return (
        "if("
        + find(d.test)
        + "){"
        + "".join(find(li) for li in d.body)
        + "}else{"
        + ("".join(find(e) for e in d.orelse) if d.orelse else "")
        + "};"
    )


@converter(ast.Call)
def call(d: ast.Call):
    args = [find(arg) for arg in d.args]
    if isinstance(d.func, ast.Name):
        if d.func.id in CALL_TYPES:
            # function call like str()
            # eg str(1) -> 1.toString()
            return f"{args[0]}.{CALL_TYPES[d.func.id]}({','.join(args[1:])})"
        elif d.func.id == "int":
            return f"parseInt({','.join(args)})"
    elif isinstance(d.func, ast.Attribute):
        value = find(d.func.value)
        # array functions like array.pop()
        if d.func.attr in CALL_REPLACEMENTS:
            return f"{value}.{CALL_REPLACEMENTS[d.func.attr]}({','.join(args)})"
        # .split .join etc
        elif d.func.attr in CALL_STRING:
            return f"{value}.{d.func.attr}({','.join(args)})"
        elif d.func.attr == "extend":
            array = find(d.func.value)
            return f"{array} = {array} + {args[0]}"
        elif (
            isinstance(d.func.value, ast.Attribute)
            and isinstance(d.func.value.value, ast.Name)
            and d.func.value.value.id == "api"
        ):
            return (
                f"API.{to_camel_case(d.func.value.attr)}.{to_camel_case(d.func.attr)}"
                "({" + dispatch_keywords(d.keywords) + "})"
            )
    msg = f"Call for {getattr(d.func, 'attr', d.func.__dict__)} is not referenced"
    raise ConverterError(msg)


@converter(ast.Pass)
def pass_expr(_: ast.Pass):
    return ""


@converter(ast.Expr)
def expr(d: ast.Expr):
    return f"{find(d.value)};"


@converter(ast.Module)
def module(d: ast.Module):
    return find(d.body)


@converter(ast.BinOp)
def bin_operation(d: ast.BinOp):
    return f"{find(d.left)}{find(d.op)}{find(d.right)}"


@converter(ast.Compare)
def compare(d: ast.Compare):
    left = find(d.left)
    operations = [
        f"{left}{find(operator)}{find(comparator)}"
        for operator, comparator in zip(d.ops, d.comparators)
    ]
    return "&&".join(operations)


@converter(ast.BoolOp)
def bool_op(d: ast.BoolOp):
    return find(d.op).join(find(value) for value in d.values)


@converter(ast.UnaryOp)
def unary_op(d: ast.UnaryOp):
    return f"{find(d.op)}{find(d.operand)}"


@converter(ast.Subscript)
def subscript(d: ast.Subscript):
    value = find(d.value)
    if d.slice.__class__ == ast.Index:
        if d.slice.value.__class__ == str:  # type: ignore  # noqa: E721
            return f"{value}.{d.slice.value.s}"  # type: ignore
        return f"{value}[{find(d.slice.value)}]"  # type: ignore
    msg = f"Slice {d.slice} is not referenced"
    raise ConverterError(msg)


@converter(ast.Attribute)
def attribute(d: ast.Attribute):
    return f"{find(d.value)}.{d.attr}"


@converter(ast.Return)
def return_statement(d: ast.Return):
    value = "null" if d is None else find(d.value)
    return f"return {value};"


@converter(ast.Delete)
def delete_statement(d: ast.Delete):
    return "".join(f"delete {find(target)};" for target in d.targets)


@converter(ast.Dict)
def dict_type(d: ast.Dict):
    dict_s = ",".join(f"{find(key)}:{find(value)}" for key, value in zip(d.keys, d.values))
    return "{" + dict_s + "}"


@converter(ast.Num)
def num_type(d: ast.Num):
    return str(d.n)


@converter(ast.Str)
def str_type(d: ast.Str):
    return repr(d.s)


@converter(ast.JoinedStr)
def joined_str(d: ast.JoinedStr):
    return "+".join(find(value) for value in d.values)


@converter(ast.FormattedValue)
def formatted_value(d: ast.FormattedValue):
    return find(d.value)


@converter(ast.List)
@converter(ast.Tuple)
def list_type(d: ast.List):
    return "[" + ",".join(find(a) for a in d.elts) + "]"


@converter(ast.NameConstant)
def name_constant_type(d: ast.NameConstant):
    consts = {True: "true", False: "false", None: "null"}
    return consts[d.value]


# Just a few typing magic
# How it works:
# func(api: ABCAPI, a: int, b: str) -> func(a: int, b: str)
P = ParamSpec("P")


def vkscript(func: Callable[Concatenate["ABCAPI", P], Any]) -> Callable[P, str]:
    def decorator(*args: P.args, **context: P.kwargs) -> str:
        return converter.scriptify(func, *args, **context)

    return decorator



from typing import Dict, Optional

from .abc import ABCStateDispenser
from .base import BaseStateGroup, StatePeer


class BuiltinStateDispenser(ABCStateDispenser):
    def __init__(self):
        self.dictionary: Dict[int, StatePeer] = {}

    async def get(self, peer_id: int) -> Optional[StatePeer]:
        return self.dictionary.get(peer_id)

    async def set(self, peer_id: int, state: BaseStateGroup, **payload):
        self.dictionary[peer_id] = StatePeer(peer_id=peer_id, state=state, payload=payload)

    async def delete(self, peer_id: int):
        self.dictionary.pop(peer_id)



from typing import TYPE_CHECKING, Union

from vkbottle.dispatch.return_manager.abc import BaseReturnManager

if TYPE_CHECKING:
    from vkbottle.tools.mini_types.user import MessageMin


class UserMessageReturnHandler(BaseReturnManager):
    @BaseReturnManager.instance_of(str)
    async def str_handler(self, value: str, message: "MessageMin", _: dict):
        await message.answer(value)

    @BaseReturnManager.instance_of((tuple, list))
    async def iter_handler(self, value: Union[tuple, list], message: "MessageMin", _: dict):
        [await message.answer(str(e)) for e in value]

    @BaseReturnManager.instance_of(dict)
    async def dict_handler(self, value: dict, message: "MessageMin", _: dict):
        await message.answer(**value)



import typing
from abc import abstractmethod
from typing import Generic, Optional, TypeVar

from .view import ABCView

if typing.TYPE_CHECKING:
    from vkbottle.dispatch.handlers.abc import ABCHandler

T_contra = TypeVar("T_contra", list, dict, contravariant=True)
F_contra = TypeVar("F_contra", contravariant=True)


class ABCDispenseView(ABCView[T_contra], Generic[T_contra, F_contra]):
    handlers: typing.List["ABCHandler[F_contra]"]

    @abstractmethod
    def get_state_key(self, event: F_contra) -> Optional[int]:
        pass



from typing import TYPE_CHECKING, Iterable

from .consistent import ConsistentTokenGenerator
from .single import SingleTokenGenerator

if TYPE_CHECKING:
    from .abc import ABCTokenGenerator, Token


def get_token_generator(token: "Token") -> "ABCTokenGenerator":
    """Converts token shortcuts to generators
    >>> get_token_generator("abc123") # <SingleTokenGenerator>
    >>> get_token_generator(["abc123", "def456"]) # <ConsistentTokenGenerator>
    >>> get_token_generator(ConsistentTokenGenerator(["abc123", "def456"])) # <ConsistentTokenGenerator>
    """
    if isinstance(token, str):
        return SingleTokenGenerator(token)
    elif isinstance(token, Iterable):
        return ConsistentTokenGenerator(token)
    return token



from .message import ABCUserMessageView, UserMessageView
from .raw import RawUserEventView, UserHandlerBasement

__all__ = (
    "ABCUserMessageView",
    "RawUserEventView",
    "UserHandlerBasement",
    "UserMessageView",
)



import re

from vkbottle.modules import pydantic

MENTION_PATTERN = re.compile(r"^\[(?P<type>club|public|id)(?P<id>\d*)\|(?P<text>.+)\],?\s?")


class Mention(pydantic.BaseModel):
    """Mention object

    :param id: Identifier of the user that was mentioned (negative if it's a group)
    :param text: Mention text
    """

    id: int
    text: str


def replace_mention_validator(cls, values):  # noqa: ARG001
    if not values.get("replace_mention"):
        return values
    message_text = values.get("text")
    if not message_text:
        return values
    match = MENTION_PATTERN.search(message_text)
    if not match:
        return values
    values["text"] = message_text.replace(match.group(0), "", 1)
    mention_id = int(match.group("id"))
    if match.group("type") in ("club", "public"):
        mention_id = -mention_id
    values["_mention"] = Mention(id=mention_id, text=match.group("text"))
    return values



from typing import Any

try:
    import contextvars
except ImportError:
    contextvars = None  # type: ignore


class BaseContext:
    """Parent BaseContext class. Idea taken from aiogram"""

    ctx_instance: Any

    def __init_subclass__(cls, **kwargs):
        if not contextvars:
            msg = f"To use {cls.__name__} you have to install contextvars"
            raise LookupError(msg)

        cls.ctx_instance = contextvars.ContextVar(kwargs.get("ctx_name") or cls.__name__)
        return cls

    @classmethod
    def get_instance(cls, no_error: bool = True) -> Any:
        if no_error:
            return cls.ctx_instance.get(None)
        return cls.ctx_instance.get()

    @classmethod
    def set_instance(cls, value: Any) -> None:
        cls.ctx_instance.set(value)


__all__ = ("BaseContext",)



from typing import Dict, Type, TypeVar

T = TypeVar("T")


def cls_kwargs(cls: Type[T], kwargs: Dict[str, object]) -> T:
    return cls(**kwargs)


class ReducibleKwargsException(Exception):
    def __reduce__(self):
        return cls_kwargs, (self.__class__, self.__dict__)



from typing import TYPE_CHECKING, Any, Callable, Union

from vkbottle.tools.magic import magic_bundle

from .abc import ABCHandler, Event

if TYPE_CHECKING:
    from vkbottle.dispatch.rules import ABCRule


class FromFuncHandler(ABCHandler[Event]):
    def __init__(self, handler: Callable, *rules: "ABCRule", blocking: bool = True):
        self.handler = handler
        self.rules = rules
        self.blocking = blocking

    async def filter(self, event: Event) -> Union[dict, bool]:
        rule_context = {}
        for rule in self.rules:
            result = await rule.check(event)
            if result is False or result is None:
                return False
            elif result is True:
                continue
            rule_context.update(result)
        return rule_context

    async def handle(self, event: Event, **context: Any) -> Any:
        return await self.handler(event, **magic_bundle(self.handler, context))

    def __eq__(self, obj: object) -> bool:
        """
        Sugar to easily check if the function is actual handler.

        >>> my_handler = lambda *args, **kwargs: None
        >>> middleware_handlers = [FromFuncHandler(my_handler)]
        >>> my_handler in middleware_handlers
        True
        """

        if callable(obj):
            return self.handler == obj
        return super().__eq__(obj)

    def __repr__(self):
        return f"<FromFuncHandler {self.handler.__name__} blocking={self.blocking} rules={self.rules}>"



from .message import ABCBotMessageView, BotMessageView
from .raw import BotHandlerBasement, RawBotEventView

__all__ = (
    "ABCBotMessageView",
    "BotHandlerBasement",
    "BotMessageView",
    "RawBotEventView",
)



import asyncio
import dataclasses
import datetime

import typing_extensions as typing

from vkbottle.api.abc import ABCAPI
from vkbottle.dispatch.handlers.abc import ABCHandler
from vkbottle.dispatch.rules.abc import ABCRule

if typing.TYPE_CHECKING:
    from .machine import Identificator

Event = typing.TypeVar("Event")
Behaviour = typing.Union[ABCHandler[Event], None]


class _ShortStateContext(typing.Generic[Event], typing.NamedTuple):
    event: Event
    context: typing.Dict[str, typing.Any]


if typing.TYPE_CHECKING:
    ShortStateContext = _ShortStateContext[Event]

else:
    ShortStateContext = typing.Annotated[_ShortStateContext[Event], ...]


@dataclasses.dataclass
class ShortState(typing.Generic[Event]):
    key: "Identificator"
    ctx_api: ABCAPI
    event: asyncio.Event
    rules: typing.Tuple[ABCRule[Event], ...]
    expiration: dataclasses.InitVar[typing.Optional[datetime.timedelta]] = dataclasses.field(
        default=None,
    )
    default_behaviour: typing.Optional[Behaviour[Event]] = dataclasses.field(default=None)
    on_drop_behaviour: typing.Optional[Behaviour[Event]] = dataclasses.field(default=None)
    exit_behaviour: typing.Optional[Behaviour[Event]] = dataclasses.field(default=None)
    expiration_date: typing.Optional[datetime.datetime] = dataclasses.field(init=False)
    context: typing.Optional[ShortStateContext[Event]] = dataclasses.field(
        default=None,
        init=False,
    )

    def __post_init__(self, expiration: typing.Optional[datetime.timedelta] = None) -> None:
        self.creation_date = datetime.datetime.now()
        self.expiration_date = (
            (self.creation_date + expiration) if expiration is not None else None
        )

    def cancel(self) -> None:
        """Cancel schedule waiters."""

        waiters = typing.cast(
            typing.Iterable[asyncio.Future[typing.Any]],
            self.event._waiters,  # type: ignore
        )
        for future in waiters:
            future.cancel()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Type

if TYPE_CHECKING:
    from types import TracebackType


class ABCHTTPClient(ABC):
    """Abstract class for http-clients
    Documentation: https://vkbottle.rtfd.io/ru/latest/low-level/http-client
    """

    @abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Any,
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Any,
    ) -> dict:
        pass

    @abstractmethod
    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Optional[dict] = None,
        **kwargs: Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "ABCHTTPClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional["TracebackType"],
    ) -> None:
        await self.close()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Dict, Type

from typing_extensions import Self

if TYPE_CHECKING:
    from vkbottle.api.abc import ABCAPI
    from vkbottle.exception_factory.error_handler import ABCErrorHandler

    from .dispenser.abc import ABCStateDispenser
    from .views import ABCView


class ABCRouter(ABC):
    """Abstract Router
    Documentation: https://vkbottle.rtfd.io/ru/latest/high-level/handling/router/
    """

    views: Dict[str, "ABCView"]
    state_dispenser: "ABCStateDispenser"
    error_handler: "ABCErrorHandler"

    def __init__(self):
        self.views = {}

    @abstractmethod
    async def route(self, event: dict, ctx_api: "ABCAPI") -> None:
        pass

    @abstractmethod
    def construct(
        self,
        views: Dict[str, "ABCView"],
        state_dispenser: "ABCStateDispenser",
        error_handler: "ABCErrorHandler",
    ) -> Self:
        pass

    def add_view(self, name: str, view: "ABCView") -> None:
        self.views[name] = view

    def view(self, name: str) -> Callable[..., Type["ABCView"]]:
        def decorator(view: Type["ABCView"]):
            self.add_view(name, view())
            return view

        return decorator



import types
import typing

Function = typing.Union[
    typing.Callable[..., typing.Any],
    types.FunctionType,
]


def resolve_arg_names(func: Function, start_idx: int = 1) -> typing.Tuple[str, ...]:
    return func.__code__.co_varnames[start_idx : func.__code__.co_argcount]


def get_default_args(func: Function) -> typing.Dict[str, typing.Any]:
    kwdefaults = func.__kwdefaults__
    if kwdefaults:
        return kwdefaults

    defaults = func.__defaults__
    if not defaults:
        return {}

    return {
        k: defaults[i]
        for i, k in enumerate(resolve_arg_names(func, start_idx=0)[-len(defaults) :])
    }


def magic_bundle(
    func: Function,
    kwargs: typing.Dict[str, typing.Any],
    *,
    start_idx: int = 1,
) -> typing.Dict[str, typing.Any]:
    names = resolve_arg_names(func, start_idx)
    args = get_default_args(func)
    args.update({k: v for k, v in kwargs.items() if k in names})
    return args


__all__ = ("resolve_arg_names", "get_default_args", "magic_bundle")



import dataclasses
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Type, TypeVar

if TYPE_CHECKING:
    from vkbottle.api.abc import ABCAPI
    from vkbottle.dispatch.dispenser.abc import ABCStateDispenser
    from vkbottle.dispatch.handlers import ABCHandler

from vkbottle.modules import logger

from .view import ABCView

Event_contra = TypeVar("Event_contra", list, dict, contravariant=True)
HandlerBasement = TypeVar("HandlerBasement", bound="BaseHandlerBasement")


@dataclasses.dataclass(frozen=True)
class BaseHandlerBasement:
    dataclass: Type[Any]
    handler: "ABCHandler"


class ABCRawEventView(ABCView[Event_contra], Generic[Event_contra, HandlerBasement]):
    handlers: Dict[Any, List[HandlerBasement]]

    @abstractmethod
    def get_handler_basements(self, event: Event_contra) -> List[HandlerBasement]:
        pass

    @abstractmethod
    def get_event_model(
        self,
        handler_basement: HandlerBasement,
        event: Event_contra,
    ) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def get_event_type(event: Event_contra) -> Any:
        pass

    async def handle_event(
        self,
        event: Event_contra,
        ctx_api: "ABCAPI",
        state_dispenser: "ABCStateDispenser",
    ) -> Any:
        logger.debug("Handling event ({}) with message view", self.get_event_type(event))

        context_variables: dict = {}
        handle_responses = []
        handlers = []

        mw_instances = await self.pre_middleware(event, context_variables)
        if mw_instances is None:
            logger.info("Handling stopped, pre_middleware returned error")
            return

        for handler_basement in self.get_handler_basements(event):
            event_model = self.get_event_model(handler_basement, event)

            if isinstance(event_model, dict):
                event_model["ctx_api"] = ctx_api
            else:
                event_model.unprepared_ctx_api = ctx_api  # type: ignore

            result = await handler_basement.handler.filter(event_model)
            logger.debug("Handler {} returned {}", handler_basement.handler, result)

            if result is False:
                continue

            elif isinstance(result, dict):
                context_variables.update(result)

            handler_response = await handler_basement.handler.handle(
                event_model, **context_variables
            )
            handle_responses.append(handler_response)
            handlers.append(handler_basement.handler)

            return_handler = self.handler_return_manager.get_handler(handler_response)
            if return_handler is not None:
                await return_handler(
                    self.handler_return_manager,
                    handler_response,
                    event_model,
                    context_variables,
                )

            if handler_basement.handler.blocking:
                break

        await self.post_middleware(mw_instances, handle_responses, handlers)



from typing import TYPE_CHECKING, List, Optional

from vkbottle_types.events.bot_events import MessageNew

from vkbottle.modules import pydantic
from vkbottle.tools.mini_types.base import BaseMessageMin

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI


from vkbottle_types.objects import ClientInfoForBots  # noqa: TCH002

from vkbottle.modules import logger

from .foreign_message import ForeignMessageMin  # noqa: TCH001


class MessageMin(BaseMessageMin):
    group_id: Optional[int] = None
    client_info: Optional["ClientInfoForBots"] = None
    reply_message: Optional["ForeignMessageMin"] = None
    fwd_messages: List["ForeignMessageMin"] = pydantic.Field(default_factory=list)
    _is_full: Optional[bool] = None

    @pydantic.root_validator(pre=True)
    def __foreign_messages(cls, values):
        foreign_messages = []
        if values.get("fwd_messages"):
            foreign_messages.extend(values["fwd_messages"])
        if values.get("reply_message"):
            foreign_messages.append(values["reply_message"])
        for foreign_message in foreign_messages:
            foreign_message["unprepared_ctx_api"] = values["unprepared_ctx_api"]
            foreign_message["replace_mention"] = values["replace_mention"]
            foreign_message["group_id"] = values["group_id"]
            foreign_message["client_info"] = values["client_info"]
        return values

    @property
    def is_mentioned(self) -> bool:
        return self.mention.id == -self.group_id if (self.mention and self.group_id) else False

    async def get_full_message(self) -> "BaseMessageMin":
        if self._is_full:
            return self
        message = (
            await self.ctx_api.messages.get_by_conversation_message_id(
                peer_id=self.peer_id,
                conversation_message_ids=[self.conversation_message_id],  # type: ignore
            )
        ).items[0]
        self.__dict__.update(message.__dict__)
        super().__init__(**self.dict())
        self.__dict__["_is_full"] = True
        if self.is_cropped:
            self.__dict__["is_cropped"] = False
        return self

    def get_attachment_strings(self) -> Optional[List[str]]:
        if self.attachments is None:
            return None
        if (
            not self.id
            and not self._is_full
            and any(
                getattr(attachment, attachment.type.value).access_key
                for attachment in self.attachments
            )
        ):
            logger.warning(
                "Some attachments may does't work because of wrong access_key. "
                "Use .get_full_message() to update message and fix this issue."
            )
        if self.is_cropped:
            logger.warning(
                "Some attachments may doesn't included because message is cropped. "
                "Use .get_full_message() to update message and fix this issue."
            )
        return super().get_attachment_strings()


def message_min(event: dict, ctx_api: "ABCAPI", replace_mention: bool = True) -> "MessageMin":
    update = MessageNew(**event)

    if update.object.message is None:
        msg = "Please set longpoll to latest version"
        raise RuntimeError(msg)

    return MessageMin(
        **update.object.message.dict(),
        client_info=update.object.client_info,
        group_id=update.group_id,
        replace_mention=replace_mention,
        unprepared_ctx_api=ctx_api,
    )


MessageMin.update_forward_refs()



from asyncio import sleep
from typing import Any, Callable, Coroutine


class DelayedTask:
    def __init__(
        self,
        seconds: int,
        handler: Callable[..., Coroutine[Any, Any, Any]],
        do_break: bool = False,
    ):
        self.seconds = seconds
        self.handler = handler
        self.do_break = do_break

    async def __call__(self, *args, **kwargs):
        while True:
            await sleep(self.seconds)
            await self.handler(*args, **kwargs)
            if self.do_break:
                break


__all__ = ("DelayedTask",)
